# Copyright (C) 2013 VMware, Inc
# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

import os_vif
from os_vif import exception as osv_exception
from oslo_concurrency import processutils
from oslo_log import log as logging

from nova import exception
from nova.i18n import _
from nova.network import linux_net
from nova.network import manager
from nova.network import model as network_model
from nova.network import os_vif_util
from nova import utils
from nova.virt.zun import network
from oslo_config import cfg

# We need config opts from manager, but pep8 complains, this silences it.
assert manager

CONF = cfg.CONF
CONF.import_opt('vlan_interface', 'nova.manager')
CONF.import_opt('flat_interface', 'nova.manager')

LOG = logging.getLogger(__name__)


class DockerGenericVIFDriver(object):

    def plug(self, instance, vif):
        vif_type = vif['type']

        LOG.debug('plug vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        if vif_type is None:
            raise exception.VirtualInterfacePlugException(
                _("vif_type parameter must be present "
                  "for this vif_driver implementation"))

        # Try os-vif codepath first
        vif_obj = os_vif_util.nova_to_osvif_vif(vif)
        if vif_obj is not None:
            self._plug_os_vif(instance, vif_obj)
            return

        # Legacy non-os-vif codepath
        func = getattr(self, 'plug_%s' % vif_type, None)
        if not func:
            raise exception.VirtualInterfacePlugException(
                _("Plug vif failed because of unexpected "
                  "vif_type=%s") % vif_type)
        func(instance, vif)

    def _plug_os_vif(self, instance, vif):
        instance_info = os_vif_util.nova_to_osvif_instance(instance)

        try:
            os_vif.plug(vif, instance_info)
        except osv_exception.ExceptionBase as ex:
            msg = (_("Failure running os_vif plugin plug method: %(ex)s")
                   % {'ex': ex})
            raise exception.NovaException(msg)

    def plug_iovisor(self, instance, vif):
        """Plug docker vif into IOvisor

        Creates a port on IOvisor and onboards the interface
        """
        if_local_name = 'tap%s' % vif['id'][:11]
        if_remote_name = 'ns%s' % vif['id'][:11]

        iface_id = vif['id']
        net_id = vif['network']['id']
        tenant_id = instance['project_id']

        # Device already exists so return.
        if linux_net.device_exists(if_local_name):
            return
        undo_mgr = utils.UndoManager()

        try:
            utils.execute('ip', 'link', 'add', 'name', if_local_name, 'type',
                          'veth', 'peer', 'name', if_remote_name,
                          run_as_root=True)
            utils.execute('ifc_ctl', 'gateway', 'add_port', if_local_name,
                          run_as_root=True)
            utils.execute('ifc_ctl', 'gateway', 'ifup', if_local_name,
                          'access_vm',
                          vif['network']['label'] + "_" + iface_id,
                          vif['address'], 'pgtag2=%s' % net_id,
                          'pgtag1=%s' % tenant_id, run_as_root=True)
            utils.execute('ip', 'link', 'set', if_local_name, 'up',
                          run_as_root=True)

        except Exception:
            LOG.exception("Failed to configure network on IOvisor")
            msg = _('Failed to setup the network, rolling back')
            undo_mgr.rollback_and_reraise(msg=msg, instance=instance)

    def plug_midonet(self, instance, vif):
        """Plug into MidoNet's network port

        This accomplishes binding of the vif to a MidoNet virtual port
        """
        if_local_name = 'tap%s' % vif['id'][:11]
        if_remote_name = 'ns%s' % vif['id'][:11]
        port_id = vif['id']

        # Device already exists so return.
        if linux_net.device_exists(if_local_name):
            return

        undo_mgr = utils.UndoManager()
        try:
            utils.execute('ip', 'link', 'add', 'name', if_local_name, 'type',
                          'veth', 'peer', 'name', if_remote_name,
                          run_as_root=True)
            undo_mgr.undo_with(lambda: utils.execute(
                'ip', 'link', 'delete', if_local_name, run_as_root=True))
            utils.execute('ip', 'link', 'set', if_local_name, 'up',
                          run_as_root=True)
            utils.execute('mm-ctl', '--bind-port', port_id, if_local_name,
                          run_as_root=True)
        except Exception:
            LOG.exception("Failed to configure network")
            msg = _('Failed to setup the network, rolling back')
            undo_mgr.rollback_and_reraise(msg=msg, instance=instance)

    # We are creating our own mac's now because the linux bridge interface
    # takes on the lowest mac that is assigned to it.  By using FE range
    # mac's we prevent the interruption and possible loss of networking
    # from changing mac addresses.
    def _fe_random_mac(self):
        mac = [0xfe, 0xed,
               random.randint(0x00, 0xff),
               random.randint(0x00, 0xff),
               random.randint(0x00, 0xff),
               random.randint(0x00, 0xff)]
        return ':'.join(map(lambda x: "%02x" % x, mac))

    def unplug(self, instance, vif):
        vif_type = vif['type']

        LOG.debug('vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        if vif_type is None:
            raise exception.NovaException(
                _("vif_type parameter must be present "
                  "for this vif_driver implementation"))

        # Try os-vif codepath first
        vif_obj = os_vif_util.nova_to_osvif_vif(vif)
        if vif_obj is not None:
            self._unplug_os_vif(instance, vif_obj)
            return

        # Legacy non-os-vif codepath
        func = getattr(self, 'unplug_%s' % vif_type, None)
        if not func:
            raise exception.NovaException(
                _("Unexpected vif_type=%s") % vif_type)
        func(instance, vif)

    def _unplug_os_vif(self, instance, vif):
        instance_info = os_vif_util.nova_to_osvif_instance(instance)

        try:
            os_vif.unplug(vif, instance_info)
        except osv_exception.ExceptionBase as ex:
            msg = (_("Failure running os_vif plugin unplug method: %(ex)s")
                   % {'ex': ex})
            raise exception.NovaException(msg)

    def unplug_iovisor(self, instance, vif):
        """Unplug vif from IOvisor

        Offboard an interface and deletes port from IOvisor
        """
        if_local_name = 'tap%s' % vif['id'][:11]
        iface_id = vif['id']
        try:
            utils.execute('ifc_ctl', 'gateway', 'ifdown',
                          if_local_name, 'access_vm',
                          vif['network']['label'] + "_" + iface_id,
                          vif['address'], run_as_root=True)
            utils.execute('ifc_ctl', 'gateway', 'del_port', if_local_name,
                          run_as_root=True)
            linux_net.delete_net_dev(if_local_name)
        except processutils.ProcessExecutionError:
            LOG.exception("Failed while unplugging vif", instance=instance)

    def unplug_midonet(self, instance, vif):
        """Unplug into MidoNet's network port

        This accomplishes unbinding of the vif from its MidoNet virtual port
        """
        try:
            utils.execute('mm-ctl', '--unbind-port',
                          vif['id'], run_as_root=True)
        except processutils.ProcessExecutionError:
            LOG.exception("Failed while unplugging vif", instance=instance)

    def _create_veth_pair(self, if_local_name, if_remote_name, bridge):
        undo_mgr = utils.UndoManager()
        try:
            utils.execute('ip', 'link', 'add', 'name', if_local_name, 'type',
                          'veth', 'peer', 'name', if_remote_name,
                          run_as_root=True)
            undo_mgr.undo_with(lambda: utils.execute(
                'ip', 'link', 'delete', if_local_name, run_as_root=True))
            # NOTE(samalba): Deleting the interface will delete all
            # associated resources (remove from the bridge, its pair, etc...)
            utils.execute('ip', 'link', 'set', if_local_name, 'address',
                          self._fe_random_mac(), run_as_root=True)
            utils.execute('brctl', 'addif', bridge, if_local_name,
                          run_as_root=True)
            utils.execute('ip', 'link', 'set', if_local_name, 'up',
                          run_as_root=True)
        except Exception:
            LOG.exception("Failed to configure network")
            msg = _('Failed to setup the network, rolling back')
            undo_mgr.rollback_and_reraise(msg=msg)

    def attach(self, instance, vif, container_id):
        vif_type = vif['type']
        if_local_name = 'tap%s' % vif['id'][:11]
        if_remote_name = 'ns%s' % vif['id'][:11]
        gateways = network.find_gateways(instance, vif['network'])
        ips = network.find_fixed_ips(instance, vif['network'])

        LOG.debug('attach vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        if not linux_net.device_exists(if_local_name):
            br_name = vif['network']['bridge']
            if (vif_type == network_model.VIF_TYPE_OVS
                    and self.ovs_hybrid_required(vif)):
                br_name = self.get_br_name(vif['id'])
            self._create_veth_pair(if_local_name, if_remote_name, br_name)

        try:
            utils.execute('ip', 'link', 'set', if_remote_name, 'netns',
                          container_id, run_as_root=True)
            utils.execute('ip', 'netns', 'exec', container_id, 'ip', 'link',
                          'set', if_remote_name, 'address', vif['address'],
                          run_as_root=True)
            for ip in ips:
                utils.execute('ip', 'netns', 'exec', container_id, 'ip',
                              'addr', 'add', ip, 'dev', if_remote_name,
                              run_as_root=True)
            utils.execute('ip', 'netns', 'exec', container_id, 'ip', 'link',
                          'set', if_remote_name, 'up', run_as_root=True)

            # Setup MTU on if_remote_name is required if it is a non
            # default value
            mtu = None
            if vif.get('mtu') is not None:
                mtu = vif.get('mtu')
            if mtu is not None:
                utils.execute('ip', 'netns', 'exec', container_id, 'ip',
                              'link', 'set', if_remote_name, 'mtu', mtu,
                              run_as_root=True)

            for gateway in gateways:
                utils.execute('ip', 'netns', 'exec', container_id,
                              'ip', 'route', 'replace', 'default', 'via',
                              gateway, 'dev', if_remote_name, run_as_root=True)

            # Disable TSO, for now no config option
            utils.execute('ip', 'netns', 'exec', container_id, 'ethtool',
                          '--offload', if_remote_name, 'tso', 'off',
                          run_as_root=True)

        except Exception:
            LOG.exception("Failed to attach vif")

    def get_br_name(self, iface_id):
        return ("qbr" + iface_id)[:network_model.NIC_NAME_LEN]

    def ovs_hybrid_required(self, vif):
        ovs_hybrid_required = self.get_firewall_required(vif) or \
            self.get_hybrid_plug_enabled(vif)
        return ovs_hybrid_required

    def get_firewall_required(self, vif):
        if vif.get('details'):
            enabled = vif['details'].get('port_filter', False)
            if enabled:
                return False
            if CONF.firewall_driver != "nova.virt.firewall.NoopFirewallDriver":
                return True
        return False

    def get_hybrid_plug_enabled(self, vif):
        if vif.get('details'):
            return vif['details'].get('ovs_hybrid_plug', False)
        return False



# Copyright 2018 NEC, Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from zun.common.policies import base

AVAILABILITY_ZONE = 'availability_zones:%s'


rules = [
    policy.DocumentedRuleDefault(
        name=AVAILABILITY_ZONE % 'get_all',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='List availability zone',
        operations=[
            {
                'path': '/v1/availability_zones',
                'method': 'GET'
            }
        ]
    )
]


def list_rules():
    return rules



# Copyright 2013 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.network import model as network_model
from zun.objects import fields as obj_fields

PCI_NET_TAG = 'physical_network'
PCI_DEVICE_TYPE_TAG = 'dev_type'

DEVICE_TYPE_FOR_VNIC_TYPE = {
    network_model.VNIC_TYPE_DIRECT_PHYSICAL: obj_fields.PciDeviceType.SRIOV_PF
}



#    Copyright 2017 ARM Holdings.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
import ssl
import websocket

from zun.common import exception
import zun.conf


CONF = zun.conf.CONF


class WebSocketClient(object):

    def __init__(self, host_url, escape='~',
                 close_wait=0.5):
        self.escape = escape
        self.close_wait = close_wait
        self.host_url = host_url
        self.cs = None

    def connect(self):
        url = self.host_url
        sslopt = None
        if url.startswith('wss'):
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_verify_locations(CONF.docker.ca_file)
            ssl_context.load_cert_chain(CONF.docker.cert_file,
                                        CONF.docker.key_file)
            sslopt = {'context': ssl_context}

        try:
            self.ws = websocket.create_connection(url,
                                                  sslopt=sslopt,
                                                  skip_utf8_validation=True)
        except socket.error as e:
            raise exception.ConnectionFailed(e)
        except websocket.WebSocketConnectionClosedException as e:
            raise exception.ConnectionFailed(e)
        except websocket.WebSocketBadStatusException as e:
            raise exception.ConnectionFailed(e)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add tty stdin_open

Revision ID: d1ef05fd92c8
Revises: ad43a2179cf2
Create Date: 2016-11-09 09:40:59.839380

"""

# revision identifiers, used by Alembic.
revision = 'd1ef05fd92c8'
down_revision = 'ad43a2179cf2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('container',
                  sa.Column('tty', sa.Boolean,
                            nullable=True))
    op.add_column('container',
                  sa.Column('stdin_open', sa.Boolean,
                            nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add started_at to containers

Revision ID: 271c7f45982d
Revises: cff60402dd86
Create Date: 2018-05-03 11:27:00.722445

"""

# revision identifiers, used by Alembic.
revision = '271c7f45982d'
down_revision = 'cff60402dd86'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('container',
                  sa.Column('started_at', sa.DateTime(),
                            nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import errno
import eventlet
import functools
import types

from docker import errors
from neutronclient.common import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils
import psutil
import tenacity

from zun.common import consts
from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
from zun.common.utils import check_container_id
from zun.compute import container_actions
import zun.conf
from zun.container.docker import host
from zun.container.docker import utils as docker_utils
from zun.container import driver
from zun.image import driver as img_driver
from zun.network import network as zun_network
from zun import objects

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)
ATTACH_FLAG = "/attach/ws?logs=0&stream=1&stdin=1&stdout=1&stderr=1"


def is_not_found(e):
    return '404' in str(e)


def is_not_connected(e):
    # Test the following exception:
    #
    #   500 Server Error: Internal Server Error ("container XXX is not
    #   connected to the network XXX")
    #
    # Note(hongbin): Docker should response a 4xx instead of 500. This looks
    # like a bug from docker side: https://github.com/moby/moby/issues/35888
    return ' is not connected to the network ' in str(e)


def is_conflict(e):
    conflict_infos = ['not running', 'not paused', 'paused']
    for info in conflict_infos:
        if info in str(e):
            return True
    return False


def handle_not_found(e, context, container, do_not_raise=False):
    if container.status == consts.DELETING:
        return

    if container.auto_remove:
        container.status = consts.DELETED
    else:
        container.status = consts.ERROR
        container.status_reason = str(e)
    container.save(context)
    if do_not_raise:
        return

    raise exception.Conflict(message=_(
        "Cannot act on container in '%s' state") % container.status)


def wrap_docker_error(function):
    @functools.wraps(function)
    def decorated_function(*args, **kwargs):
        context = args[1]
        container = args[2]
        try:
            return function(*args, **kwargs)
        except exception.DockerError as e:
            if is_not_found(e):
                handle_not_found(e, context, container)
            if is_conflict(e):
                raise exception.Conflict(_("%s") % str(e))
            raise

    return decorated_function


class DockerDriver(driver.BaseDriver, driver.ContainerDriver,
                   driver.CapsuleDriver):
    """Implementation of container drivers for Docker."""

    # TODO(hongbin): define a list of capabilities of this driver.
    capabilities = {}

    def __init__(self):
        super(DockerDriver, self).__init__()
        self._host = host.Host()
        self._get_host_storage_info()
        self.image_drivers = {}
        for driver_name in CONF.image_driver_list:
            driver = img_driver.load_image_driver(driver_name)
            self.image_drivers[driver_name] = driver

    def _get_host_storage_info(self):
        host_info = self.get_host_info()
        self.docker_root_dir = host_info['docker_root_dir']
        storage_info = self._host.get_storage_info()
        self.base_device_size = storage_info['default_base_size']
        self.support_disk_quota = self._host.check_supported_disk_quota(
            host_info)

    def load_image(self, image_path=None):
        with docker_utils.docker_client() as docker:
            if image_path:
                with open(image_path, 'rb') as fd:
                    LOG.debug('Loading local image %s into docker', image_path)
                    docker.load_image(fd)

    def inspect_image(self, image):
        with docker_utils.docker_client() as docker:
            LOG.debug('Inspecting image %s', image)
            return docker.inspect_image(image)

    def get_image(self, name):
        LOG.debug('Obtaining image %s', name)
        with docker_utils.docker_client() as docker:
            return docker.get_image(name)

    def delete_image(self, context, img_id, image_driver=None):
        image = self.inspect_image(img_id)['RepoTags'][0]
        if image_driver:
            image_driver_list = [image_driver.lower()]
        else:
            image_driver_list = CONF.image_driver_list
        for driver_name in image_driver_list:
            try:
                image_driver = img_driver.load_image_driver(driver_name)
                if driver_name == 'glance':
                    image_driver.delete_image_tar(context, image)
                elif driver_name == 'docker':
                    image_driver.delete_image(context, img_id)
            except exception.ZunException:
                LOG.exception('Unknown exception occurred while deleting '
                              'image %s', img_id)

    def delete_committed_image(self, context, img_id, image_driver):
        try:
            image_driver.delete_committed_image(context, img_id)
        except Exception as e:
            LOG.exception('Unknown exception occurred while '
                          'deleting image %s: %s',
                          img_id,
                          str(e))
            raise exception.ZunException(str(e))

    def images(self, repo, quiet=False):
        with docker_utils.docker_client() as docker:
            return docker.images(repo, quiet)

    def pull_image(self, context, repo, tag, image_pull_policy='always',
                   driver_name=None, registry=None):
        if driver_name is None:
            driver_name = CONF.default_image_driver

        try:
            image_driver = self.image_drivers[driver_name]
            image, image_loaded = image_driver.pull_image(
                context, repo, tag, image_pull_policy, registry)
            if image:
                image['driver'] = driver_name.split('.')[0]
        except exception.ZunException:
            raise
        except Exception as e:
            LOG.exception('Unknown exception occurred while loading '
                          'image: %s', str(e))
            raise exception.ZunException(str(e))

        return image, image_loaded

    def search_image(self, context, repo, tag, driver_name, exact_match):
        if driver_name is None:
            driver_name = CONF.default_image_driver

        try:
            image_driver = self.image_drivers[driver_name]
            return image_driver.search_image(context, repo, tag,
                                             exact_match)
        except exception.ZunException:
            raise
        except Exception as e:
            LOG.exception('Unknown exception occurred while searching '
                          'for image: %s', str(e))
            raise exception.ZunException(str(e))

    def create_image(self, context, image_name, image_driver):
        try:
            img = image_driver.create_image(context, image_name)
        except Exception as e:
            LOG.exception('Unknown exception occurred while creating '
                          'image: %s', str(e))
            raise exception.ZunException(str(e))
        return img

    def upload_image_data(self, context, image, image_tag, image_data,
                          image_driver):
        try:
            image_driver.update_image(context,
                                      image.id,
                                      tag=image_tag)
            # Image data has to match the image format.
            # contain format defaults to 'docker';
            # disk format defaults to 'qcow2'.
            img = image_driver.upload_image_data(context,
                                                 image.id,
                                                 image_data)
        except Exception as e:
            LOG.exception('Unknown exception occurred while uploading '
                          'image: %s', str(e))
            raise exception.ZunException(str(e))
        return img

    def read_tar_image(self, image):
        with docker_utils.docker_client() as docker:
            LOG.debug('Reading local tar image %s ', image['path'])
            try:
                docker.read_tar_image(image)
            except Exception:
                LOG.warning("Unable to read image data from tarfile")

    def create(self, context, container, image, requested_networks,
               requested_volumes):
        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            name = container.name
            if image['tag']:
                image_repo = image['repo'] + ":" + image['tag']
            else:
                image_repo = image['repo']
            LOG.debug('Creating container with image %(image)s name %(name)s',
                      {'image': image_repo, 'name': name})
            self._provision_network(context, network_driver,
                                    requested_networks)
            volmaps = requested_volumes.get(container.uuid, [])
            binds = self._get_binds(context, volmaps)

            entrypoint = container.entrypoint
            command = container.command
            if not entrypoint or not command:
                image_dict = docker.inspect_image(image_repo)
                container.entrypoint = entrypoint or \
                    image_dict['Config']['Entrypoint']
                container.command = command or image_dict['Config']['Cmd']
            kwargs = {
                'name': self.get_container_name(container),
                'command': container.command,
                'environment': container.environment,
                'working_dir': container.workdir,
                'labels': container.labels,
                'tty': container.tty,
                'stdin_open': container.interactive,
                'hostname': container.hostname,
                'entrypoint': container.entrypoint,
            }

            if not self._is_runtime_supported():
                if container.runtime:
                    raise exception.ZunException(_(
                        'Specifying runtime in Docker API is not supported'))
                runtime = None
            else:
                runtime = container.runtime or CONF.container_runtime

            host_config = {}
            host_config['privileged'] = container.privileged
            host_config['runtime'] = runtime
            host_config['binds'] = binds
            kwargs['volumes'] = [b['bind'] for b in binds.values()]
            self._process_exposed_ports(network_driver.neutron_api, container,
                                        kwargs)
            # Process the first requested network at create time. The rest
            # will be processed after create.
            requested_network = requested_networks.pop()
            security_group_ids = utils.get_security_group_ids(
                context, container.security_groups)
            network_driver.process_networking_config(
                container, requested_network, host_config, kwargs, docker,
                security_group_ids=security_group_ids)
            if container.auto_remove:
                host_config['auto_remove'] = container.auto_remove
            if self._should_limit_memory(container):
                host_config['mem_limit'] = str(container.memory) + 'M'
            if self._should_limit_cpu(container):
                host_config['cpu_shares'] = int(1024 * container.cpu)
            if container.restart_policy:
                count = int(container.restart_policy['MaximumRetryCount'])
                name = container.restart_policy['Name']
                host_config['restart_policy'] = {'Name': name,
                                                 'MaximumRetryCount': count}

            if container.disk:
                disk_size = str(container.disk) + 'G'
                host_config['storage_opt'] = {'size': disk_size}
            if container.cpu_policy == 'dedicated':
                host_config['cpuset_cpus'] = container.cpuset.cpuset_cpus
                host_config['cpuset_mems'] = str(container.cpuset.cpuset_mems)
            # The time unit in docker of heath checking is us, and the unit
            # of interval and timeout is seconds.
            if container.healthcheck:
                healthcheck = {}
                healthcheck['test'] = container.healthcheck.get('test', '')
                interval = container.healthcheck.get('interval', 0)
                healthcheck['interval'] = interval * 10 ** 9
                healthcheck['retries'] = int(container.healthcheck.
                                             get('retries', 0))
                timeout = container.healthcheck.get('timeout', 0)
                healthcheck['timeout'] = timeout * 10 ** 9
                kwargs['healthcheck'] = healthcheck

            kwargs['host_config'] = docker.create_host_config(**host_config)
            response = docker.create_container(image_repo, **kwargs)
            container.container_id = response['Id']

            addresses = self._setup_network_for_container(
                context, container, requested_networks, network_driver)
            container.addresses = addresses

            response = docker.inspect_container(container.container_id)
            self._populate_container(container, response, force=True)
            container.save(context)
            return container

    def _should_limit_memory(self, container):
        return (container.memory is not None and
                not isinstance(container, objects.Capsule))

    def _should_limit_cpu(self, container):
        return (container.cpu is not None and
                not isinstance(container, objects.Capsule))

    def _is_runtime_supported(self):
        return float(CONF.docker.docker_remote_api_version) >= 1.26

    def node_support_disk_quota(self):
        return self.support_disk_quota

    def get_host_default_base_size(self):
        return self.base_device_size

    def _process_exposed_ports(self, neutron_api, container, kwargs):
        exposed_ports = {}
        if isinstance(container, objects.Container):
            exposed_ports.update(container.exposed_ports or {})
        if isinstance(container, objects.Capsule):
            for container in (container.init_containers +
                              container.containers):
                exposed_ports.update(container.exposed_ports or {})

        if not exposed_ports:
            return

        # process security group
        secgroup_name = self._get_secgorup_name(container.uuid)
        secgroup_id = neutron_api.create_security_group({'security_group': {
            "name": secgroup_name}})['security_group']['id']
        neutron_api.expose_ports(secgroup_id, exposed_ports)
        if container.security_groups:
            container.security_groups.append(secgroup_id)
        else:
            container.security_groups = [secgroup_id]
        # process kwargs on creating the docker container
        ports = []
        for port in exposed_ports:
            port, proto = port.split('/')
            ports.append((port, proto))
        kwargs['ports'] = ports

    def _provision_network(self, context, network_driver, requested_networks):
        for rq_network in requested_networks:
            network_driver.get_or_create_network(context,
                                                 rq_network['network'])

    def _get_secgorup_name(self, container_uuid):
        return consts.NAME_PREFIX + container_uuid

    def _get_binds(self, context, requested_volumes):
        binds = {}
        for volume in requested_volumes:
            volume_driver = self._get_volume_driver(volume)
            source, destination = volume_driver.bind_mount(context, volume)
            binds[source] = {'bind': destination}
        return binds

    def _setup_network_for_container(self, context, container,
                                     requested_networks, network_driver):
        security_group_ids = utils.get_security_group_ids(
            context, container.security_groups)
        addresses = {}
        if container.addresses:
            addresses = container.addresses
        for network in requested_networks:
            if network['network'] in addresses:
                # This network is already setup so skip it
                continue

            addrs = network_driver.connect_container_to_network(
                container, network, security_groups=security_group_ids)
            addresses[network['network']] = addrs

        return addresses

    def delete(self, context, container, force):
        with docker_utils.docker_client() as docker:
            try:
                network_driver = zun_network.driver(context=context,
                                                    docker_api=docker)
                self._cleanup_network_for_container(container, network_driver)
                self._cleanup_exposed_ports(network_driver.neutron_api,
                                            container)
                if container.container_id:
                    docker.remove_container(container.container_id,
                                            force=force)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    return
                if is_not_connected(api_error):
                    return
                raise

    @wrap_docker_error
    def _cleanup_network_for_container(self, container, network_driver):
        if not container.addresses:
            return
        for neutron_net in container.addresses:
            network_driver.disconnect_container_from_network(
                container, neutron_net)

    def _cleanup_exposed_ports(self, neutron_api, container):
        exposed_ports = {}
        if isinstance(container, objects.Container):
            exposed_ports.update(container.exposed_ports or {})
        if isinstance(container, objects.Capsule):
            for container in (container.init_containers +
                              container.containers):
                exposed_ports.update(container.exposed_ports or {})
        if not exposed_ports:
            return

        try:
            neutron_api.delete_security_group(container.security_groups[0])
        except n_exc.NeutronClientException:
            LOG.exception("Failed to delete security group")

    def check_container_exist(self, container):
        with docker_utils.docker_client() as docker:
            docker_containers = [c['Id']
                                 for c in docker.list_containers()]
            if container.container_id not in docker_containers:
                return False
        return True

    def list(self, context):
        non_existent_containers = []
        with docker_utils.docker_client() as docker:
            docker_containers = docker.list_containers()
            id_to_container_map = {c['Id']: c
                                   for c in docker_containers}
            uuids = self._get_container_uuids(docker_containers)

        local_containers = self._get_local_containers(context, uuids)
        for container in local_containers:
            if container.status in (consts.CREATING, consts.DELETING,
                                    consts.DELETED):
                # Skip populating db record since the container is in a
                # unstable state.
                continue

            container_id = container.container_id
            docker_container = id_to_container_map.get(container_id)
            if not container_id or not docker_container:
                non_existent_containers.append(container)
                continue

            self._populate_container(container, docker_container)

        return local_containers, non_existent_containers

    def heal_with_rebuilding_container(self, context, container, manager):
        if not container.container_id:
            return

        rebuild_status = utils.VALID_STATES['rebuild']
        try:
            if (container.auto_heal and
                    container.status in rebuild_status):
                context.project_id = container.project_id
                objects.ContainerAction.action_start(
                    context, container.uuid, container_actions.REBUILD,
                    want_result=False)
                manager.container_rebuild(context, container)
            else:
                LOG.warning("Container %s was recorded in DB but "
                            "missing in docker", container.uuid)
                container.status = consts.ERROR
                msg = "No such container: %s in docker" % \
                      (container.container_id)
                container.status_reason = str(msg)
                container.save(context)
        except Exception as e:
            LOG.warning("heal container with rebuilding failed, "
                        "err code: %s", e)

    def _get_container_uuids(self, containers):
        # The name of Docker container is of the form '/zun-<uuid>'
        name_prefix = '/' + consts.NAME_PREFIX
        uuids = [c['Names'][0].replace(name_prefix, '', 1)
                 for c in containers]
        return [u for u in uuids if uuidutils.is_uuid_like(u)]

    def _get_local_containers(self, context, uuids):
        host_containers = objects.Container.list_by_host(context, CONF.host)
        uuids = list(set(uuids) | set([c.uuid for c in host_containers]))
        containers = objects.Container.list(context,
                                            filters={'uuid': uuids})
        return containers

    def update_containers_states(self, context, containers, manager):
        local_containers, non_existent_containers = self.list(context)
        if not local_containers:
            return

        id_to_local_container_map = {container.container_id: container
                                     for container in local_containers
                                     if container.container_id}
        id_to_container_map = {container.container_id: container
                               for container in containers
                               if container.container_id}

        for cid in (id_to_container_map.keys() &
                    id_to_local_container_map.keys()):
            container = id_to_container_map[cid]
            # sync status
            local_container = id_to_local_container_map[cid]
            if container.status != local_container.status:
                old_status = container.status
                container.status = local_container.status
                container.save(context)
                LOG.info('Status of container %s changed from %s to %s',
                         container.uuid, old_status, container.status)
            # sync host
            # Note(kiennt): Current host.
            cur_host = CONF.host
            if container.host != cur_host:
                old_host = container.host
                container.host = cur_host
                container.save(context)
                LOG.info('Host of container %s changed from %s to %s',
                         container.uuid, old_host, container.host)
        for container in non_existent_containers:
            if container.host == CONF.host:
                if container.auto_remove:
                    container.status = consts.DELETED
                    container.save(context)
                else:
                    self.heal_with_rebuilding_container(context, container,
                                                        manager)

    def show(self, context, container):
        with docker_utils.docker_client() as docker:
            if container.container_id is None:
                return container

            response = None
            try:
                response = docker.inspect_container(container.container_id)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    handle_not_found(api_error, context, container,
                                     do_not_raise=True)
                    return container
                raise

            self._populate_container(container, response)
            return container

    def format_status_detail(self, status_time):
        try:
            st = datetime.datetime.strptime((status_time[:19]),
                                            '%Y-%m-%dT%H:%M:%S')
        except ValueError as e:
            LOG.exception("Error on parse {} : {}".format(status_time, e))
            return

        if st == datetime.datetime(1, 1, 1):
            # return empty string if the time is January 1, year 1, 00:00:00
            return ""

        delta = timeutils.utcnow() - st
        time_dict = {}
        time_dict['days'] = delta.days
        time_dict['hours'] = delta.seconds // 3600
        time_dict['minutes'] = (delta.seconds % 3600) // 60
        time_dict['seconds'] = delta.seconds
        if time_dict['days']:
            return '{} days'.format(time_dict['days'])
        if time_dict['hours']:
            return '{} hours'.format(time_dict['hours'])
        if time_dict['minutes']:
            return '{} mins'.format(time_dict['minutes'])
        if time_dict['seconds']:
            return '{} seconds'.format(time_dict['seconds'])
        return

    def _populate_container(self, container, response, force=False):
        state = response.get('State')
        self._populate_container_state(container, state, force)

        config = response.get('Config')
        if config:
            self._populate_hostname_and_ports(container, config)

        hostconfig = response.get('HostConfig')
        if hostconfig:
            container.runtime = hostconfig.get('Runtime')

    def _populate_container_state(self, container, state, force):
        if container.task_state and not force:
            # NOTE(hongbin): we don't want to populate container state
            # if another thread is performing task on this container.
            # In this case, task_state will be assigned (which means there is a
            # task performing on this container) and 'force' will be set to
            # False. For example, if this method is called by create,
            # 'force' will be set to True to force refreshing the state.
            # If this method is called by list or show,
            # 'force' will be set to False, in which case we skip
            # refreshing the state if there is a task on this container.
            return

        if not state:
            LOG.warning('Receive unexpected state from docker: %s', state)
            container.status = consts.UNKNOWN
            container.status_reason = _("container state is missing")
            container.status_detail = None
        elif type(state) is dict:
            status_detail = ''
            if state.get('Error'):
                if state.get('Status') in ('exited', 'removing'):
                    container.status = consts.STOPPED
                else:
                    status = state.get('Status').capitalize()
                    if status in consts.CONTAINER_STATUSES:
                        container.status = status
                status_detail = self.format_status_detail(
                    state.get('FinishedAt'))
                container.status_detail = "Exited({}) {} ago (error)".format(
                    state.get('ExitCode'), status_detail)
            elif state.get('Paused'):
                container.status = consts.PAUSED
                status_detail = self.format_status_detail(
                    state.get('StartedAt'))
                container.status_detail = "Up {} (paused)".format(
                    status_detail)
            elif state.get('Restarting'):
                container.status = consts.RESTARTING
                container.status_detail = "Restarting"
            elif state.get('Running'):
                container.status = consts.RUNNING
                status_detail = self.format_status_detail(
                    state.get('StartedAt'))
                container.status_detail = "Up {}".format(
                    status_detail)
            elif state.get('Dead'):
                container.status = consts.DEAD
                container.status_detail = "Dead"
            else:
                started_at = self.format_status_detail(state.get('StartedAt'))
                finished_at = self.format_status_detail(
                    state.get('FinishedAt'))
                if started_at == "" and container.status == consts.CREATING:
                    container.status = consts.CREATED
                    container.status_detail = "Created"
                elif (started_at == "" and
                      container.status in (consts.CREATED, consts.RESTARTING,
                                           consts.ERROR, consts.REBUILDING)):
                    pass
                elif started_at != "" and finished_at == "":
                    LOG.warning('Receive unexpected state from docker: %s',
                                state)
                    container.status = consts.UNKNOWN
                    container.status_reason = _("unexpected container state")
                    container.status_detail = ""
                elif started_at != "" and finished_at != "":
                    container.status = consts.STOPPED
                    container.status_detail = "Exited({}) {} ago ".format(
                        state.get('ExitCode'), finished_at)
            if status_detail is None:
                container.status_detail = None
        else:
            state = state.lower()
            if state == 'created' and container.status == consts.CREATING:
                container.status = consts.CREATED
            elif (state == 'created' and
                  container.status in (consts.CREATED, consts.RESTARTING,
                                       consts.ERROR, consts.REBUILDING)):
                pass
            elif state == 'paused':
                container.status = consts.PAUSED
            elif state == 'running':
                container.status = consts.RUNNING
            elif state == 'dead':
                container.status = consts.DEAD
            elif state == 'restarting':
                container.status = consts.RESTARTING
            elif state in ('exited', 'removing'):
                container.status = consts.STOPPED
            else:
                LOG.warning('Receive unexpected state from docker: %s', state)
                container.status = consts.UNKNOWN
                container.status_reason = _("unexpected container state")
            container.status_detail = None

    def _populate_hostname_and_ports(self, container, config):
        # populate hostname only when container.hostname wasn't set
        if container.hostname is None:
            container.hostname = config.get('Hostname')
        # populate ports
        ports = []
        exposed_ports = config.get('ExposedPorts')
        if exposed_ports:
            for key in exposed_ports:
                port = key.split('/')[0]
                ports.append(int(port))
        container.ports = ports

    @check_container_id
    @wrap_docker_error
    def reboot(self, context, container, timeout):
        with docker_utils.docker_client() as docker:
            if timeout:
                docker.restart(container.container_id,
                               timeout=int(timeout))
            else:
                docker.restart(container.container_id)
            container.status = consts.RUNNING
            container.status_reason = None
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            network_driver.on_container_started(container)
            return container

    @check_container_id
    @wrap_docker_error
    def stop(self, context, container, timeout):
        with docker_utils.docker_client() as docker:
            if timeout:
                docker.stop(container.container_id,
                            timeout=int(timeout))
            else:
                docker.stop(container.container_id)
            container.status = consts.STOPPED
            container.status_reason = None
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            try:
                network_driver.on_container_stopped(container)
            except Exception as e:
                LOG.error('network driver failed on stopping container: %s',
                          str(e))
            return container

    @check_container_id
    @wrap_docker_error
    def start(self, context, container):
        with docker_utils.docker_client() as docker:
            docker.start(container.container_id)
            container.status = consts.RUNNING
            container.status_reason = None
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            try:
                network_driver.on_container_started(container)
            except Exception as e:
                LOG.error('network driver failed on starting container: %s',
                          str(e))
                try:
                    docker.stop(container.container_id, timeout=5)
                    LOG.debug('Stop container successfully')
                    network_driver.on_container_stopped(container)
                    LOG.debug('Network driver clean up successfully')
                except Exception:
                    pass
                container.status = consts.STOPPED
                container.status_reason = _("failed to configure network")
            return container

    @check_container_id
    @wrap_docker_error
    def pause(self, context, container):
        with docker_utils.docker_client() as docker:
            docker.pause(container.container_id)
            container.status = consts.PAUSED
            container.status_reason = None
            return container

    @check_container_id
    @wrap_docker_error
    def unpause(self, context, container):
        with docker_utils.docker_client() as docker:
            docker.unpause(container.container_id)
            container.status = consts.RUNNING
            container.status_reason = None
            return container

    @check_container_id
    @wrap_docker_error
    def show_logs(self, context, container, stdout=True, stderr=True,
                  timestamps=False, tail='all', since=None):
        with docker_utils.docker_client() as docker:
            try:
                tail = int(tail)
            except ValueError:
                tail = 'all'

            if since is None or since == 'None':
                return docker.logs(container.container_id, stdout, stderr,
                                   False, timestamps, tail, None)
            else:
                try:
                    since = int(since)
                except ValueError:
                    try:
                        since = datetime.datetime.strptime(
                            since, '%Y-%m-%d %H:%M:%S,%f')
                    except Exception:
                        raise
                return docker.logs(container.container_id, stdout, stderr,
                                   False, timestamps, tail, since)

    @check_container_id
    @wrap_docker_error
    def execute_create(self, context, container, command, interactive=False):
        stdin = True if interactive else False
        tty = True if interactive else False
        with docker_utils.docker_client() as docker:
            create_res = docker.exec_create(
                container.container_id, command, stdin=stdin, tty=tty)
            exec_id = create_res['Id']
            return exec_id

    def execute_run(self, exec_id, command):
        with docker_utils.docker_client() as docker:
            try:
                with eventlet.Timeout(CONF.docker.execute_timeout):
                    output = docker.exec_start(exec_id, False, False, False)
            except eventlet.Timeout:
                raise exception.Conflict(_(
                    "Timeout on executing command: %s") % command)
            inspect_res = docker.exec_inspect(exec_id)
            return output, inspect_res['ExitCode']

    def execute_resize(self, exec_id, height, width):
        height = int(height)
        width = int(width)
        with docker_utils.docker_client() as docker:
            try:
                docker.exec_resize(exec_id, height=height, width=width)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    raise exception.Invalid(_(
                        "no such exec instance: %s") % str(api_error))
                raise

    @check_container_id
    @wrap_docker_error
    def kill(self, context, container, signal=None):
        with docker_utils.docker_client() as docker:
            if signal is None or signal == 'None':
                docker.kill(container.container_id)
            else:
                docker.kill(container.container_id, signal)
            container.status = consts.STOPPED
            container.status_reason = None
            return container

    @check_container_id
    @wrap_docker_error
    def update(self, context, container):
        patch = container.obj_get_changes()

        args = {}
        memory = patch.get('memory')
        if memory is not None:
            args['mem_limit'] = str(memory) + 'M'
            args['memswap_limit'] = CONF.default_memory_swap
        cpu = patch.get('cpu')
        if cpu is not None:
            args['cpu_shares'] = int(1024 * cpu)

        with docker_utils.docker_client() as docker:
            return docker.update_container(container.container_id, **args)

    @check_container_id
    def get_websocket_url(self, context, container):
        protocol = "wss" if (not CONF.docker.api_insecure and
                             CONF.docker.ca_file and
                             CONF.docker.key_file and
                             CONF.docker.cert_file) else "ws"
        version = CONF.docker.docker_remote_api_version
        remote_api_host = CONF.docker.docker_remote_api_host
        remote_api_port = CONF.docker.docker_remote_api_port
        url = protocol + "://" + remote_api_host + ":" + remote_api_port \
            + "/v" + version + "/containers/" + container.container_id \
            + ATTACH_FLAG
        return url

    @check_container_id
    @wrap_docker_error
    def resize(self, context, container, height, width):
        with docker_utils.docker_client() as docker:
            height = int(height)
            width = int(width)
            docker.resize(container.container_id, height, width)
            return container

    @check_container_id
    @wrap_docker_error
    def top(self, context, container, ps_args=None):
        with docker_utils.docker_client() as docker:
            if ps_args is None or ps_args == 'None':
                return docker.top(container.container_id)
            else:
                return docker.top(container.container_id, ps_args)

    @check_container_id
    @wrap_docker_error
    def get_archive(self, context, container, path):
        with docker_utils.docker_client() as docker:
            try:
                stream, stat = docker.get_archive(
                    container.container_id, path)
                if isinstance(stream, types.GeneratorType):
                    filedata = ''.encode("latin-1").join(stream)
                else:
                    filedata = stream.read()
                return filedata, stat
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    raise exception.Invalid(_("%s") % str(api_error))
                raise

    @check_container_id
    @wrap_docker_error
    def put_archive(self, context, container, path, data):
        with docker_utils.docker_client() as docker:
            try:
                docker.put_archive(container.container_id, path, data)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    raise exception.Invalid(_("%s") % str(api_error))
                raise

    @check_container_id
    @wrap_docker_error
    def stats(self, context, container):
        with docker_utils.docker_client() as docker:
            res = docker.stats(container.container_id, decode=False,
                               stream=False)

            cpu_usage = res['cpu_stats']['cpu_usage']['total_usage']
            system_cpu_usage = res['cpu_stats']['system_cpu_usage']
            cpu_percent = float(cpu_usage) / float(system_cpu_usage) * 100

            # Subtract the Cache Usage from total memory Usage
            cache_usage = 0
            if res['memory_stats'].get('stats'):
                if res['memory_stats']['stats'].get('total_inactive_file'):
                    # For Cgroup V1
                    cache_usage = res['memory_stats']['stats'].get(
                        'total_inactive_file')
                elif res['memory_stats']['stats'].get('inactive_file'):
                    # For Cgroup V2
                    cache_usage = res['memory_stats']['stats'].get(
                        'inactive_file')

            mem_usage = res['memory_stats']['usage'] - cache_usage
            mem_usage = mem_usage / 1024 / 1024
            mem_limit = res['memory_stats']['limit'] / 1024 / 1024
            mem_percent = float(mem_usage) / float(mem_limit) * 100

            blk_stats = res['blkio_stats']['io_service_bytes_recursive']
            io_read = 0
            io_write = 0
            for item in (blk_stats or []):
                if item['op'].lower() == 'read':
                    io_read = io_read + item['value']
                if item['op'].lower() == 'write':
                    io_write = io_write + item['value']

            # Note(hongbin): CNI network won't have this key
            net_stats = res.get('networks', {})
            net_rxb = 0
            net_txb = 0
            for k, v in net_stats.items():
                net_rxb = net_rxb + v['rx_bytes']
                net_txb = net_txb + v['tx_bytes']

            stats = {"CONTAINER": container.name,
                     "CPU %": cpu_percent,
                     "MEM USAGE(MiB)": mem_usage,
                     "MEM LIMIT(MiB)": mem_limit,
                     "MEM %": mem_percent,
                     "BLOCK I/O(B)": str(io_read) + "/" + str(io_write),
                     "NET I/O(B)": str(net_rxb) + "/" + str(net_txb)}
            return stats

    @check_container_id
    @wrap_docker_error
    def commit(self, context, container, repository=None, tag=None):
        with docker_utils.docker_client() as docker:
            repository = str(repository)
            if tag is None or tag == "None":
                return docker.commit(container.container_id, repository)
            else:
                tag = str(tag)
                return docker.commit(container.container_id, repository, tag)

    def _encode_utf8(self, value):
        return value.encode('utf-8')

    def get_container_name(self, container):
        return consts.NAME_PREFIX + container.uuid

    def get_host_info(self):
        with docker_utils.docker_client() as docker:
            info = docker.info()
            total = info['Containers']
            paused = info['ContainersPaused']
            running = info['ContainersRunning']
            stopped = info['ContainersStopped']
            cpus = info['NCPU']
            architecture = info['Architecture']
            os_type = info['OSType']
            os = info['OperatingSystem']
            kernel_version = info['KernelVersion']
            labels = {}
            slabels = info['Labels']
            if slabels:
                for slabel in slabels:
                    kv = slabel.split("=")
                    label = {kv[0]: kv[1]}
                    labels.update(label)
            runtimes = []
            if 'Runtimes' in info:
                for key in info['Runtimes']:
                    runtimes.append(key)
            else:
                runtimes = ['runc']
            docker_root_dir = info['DockerRootDir']

            return {'total_containers': total,
                    'running_containers': running,
                    'paused_containers': paused,
                    'stopped_containers': stopped,
                    'cpus': cpus,
                    'architecture': architecture,
                    'os_type': os_type,
                    'os': os,
                    'kernel_version': kernel_version,
                    'labels': labels,
                    'runtimes': runtimes,
                    'docker_root_dir': docker_root_dir}

    def get_total_disk_for_container(self):
        try:
            disk_usage = psutil.disk_usage(self.docker_root_dir)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            LOG.warning('Docker data root doesnot exist.')
            # give another try with system root
            disk_usage = psutil.disk_usage('/')
        total_disk = disk_usage.total / 1024 ** 3
        # TODO(hongbin): deprecate reserve_disk_for_image in flavor of
        # reserved_host_disk_mb
        return (int(total_disk),
                int(total_disk * CONF.compute.reserve_disk_for_image))

    def add_security_group(self, context, container, security_group):

        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            network_driver.add_security_groups_to_ports(container,
                                                        [security_group])

    def remove_security_group(self, context, container, security_group):

        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context=context,
                                                docker_api=docker)
            network_driver.remove_security_groups_from_ports(container,
                                                             [security_group])

    def get_available_nodes(self):
        return [self._host.get_hostname()]

    def get_available_resources(self):
        data = super(DockerDriver, self).get_available_resources()

        info = self.get_host_info()
        data['total_containers'] = info['total_containers']
        data['running_containers'] = info['running_containers']
        data['paused_containers'] = info['paused_containers']
        data['stopped_containers'] = info['stopped_containers']
        data['cpus'] = info['cpus']
        data['architecture'] = info['architecture']
        data['os_type'] = info['os_type']
        data['os'] = info['os']
        data['kernel_version'] = info['kernel_version']
        data['labels'] = info['labels']
        data['runtimes'] = info['runtimes']

        return data

    @wrap_docker_error
    def network_detach(self, context, container, network):
        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context,
                                                docker_api=docker)
            network_driver.disconnect_container_from_network(container,
                                                             network)

            # Only clear network info related to this network
            # Cannot del container.address directly which will not update
            # changed fields of the container objects as the del operate on
            # the addresses object, only base.getter will called.
            update = container.addresses
            del update[network]
            container.addresses = update
            container.save(context)

    def network_attach(self, context, container, requested_network):
        with docker_utils.docker_client() as docker:
            security_group_ids = None
            if container.security_groups:
                security_group_ids = utils.get_security_group_ids(
                    context, container.security_groups)
            network_driver = zun_network.driver(context, docker_api=docker)
            network = requested_network['network']
            if network in container.addresses:
                raise exception.ZunException('Container %(container)s has '
                                             'already connected to the '
                                             'network %(network)s.'
                                             % {'container': container.uuid,
                                                'network': network})
            network_driver.get_or_create_network(context, network)
            addrs = network_driver.connect_container_to_network(
                container, requested_network,
                security_groups=security_group_ids)
            if addrs is None:
                raise exception.ZunException(_(
                    'Unexpected missing of addresses'))
            update = {}
            update[network] = addrs
            addresses = container.addresses
            addresses.update(update)
            container.addresses = addresses
            container.save(context)

    def create_network(self, context, neutron_net_id):
        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context, docker_api=docker)
            return network_driver.create_network(neutron_net_id)

    def delete_network(self, context, network):
        with docker_utils.docker_client() as docker:
            network_driver = zun_network.driver(context,
                                                docker_api=docker)
            network_driver.remove_network(network)

    def create_capsule(self, context, capsule, image, requested_networks,
                       requested_volumes):
        capsule = self.create(context, capsule, image, requested_networks,
                              requested_volumes)
        self.start(context, capsule)
        for container in capsule.init_containers:
            self._create_container_in_capsule(context, capsule, container,
                                              requested_networks,
                                              requested_volumes)
            self._wait_for_init_container(context, container)
            container.save(context)
        for container in capsule.containers:
            self._create_container_in_capsule(context, capsule, container,
                                              requested_networks,
                                              requested_volumes)
        return capsule

    def _create_container_in_capsule(self, context, capsule, container,
                                     requested_networks, requested_volumes):
        # pull image
        image_driver_name = container.image_driver
        repo, tag = utils.parse_image_name(container.image, image_driver_name)
        image_pull_policy = utils.get_image_pull_policy(
            container.image_pull_policy, tag)
        image, image_loaded = self.pull_image(
            context, repo, tag, image_pull_policy, image_driver_name)
        image['repo'], image['tag'] = repo, tag
        if not image_loaded:
            self.load_image(image['path'])
        if image_driver_name == 'glance':
            self.read_tar_image(image)
        if image['tag'] != tag:
            LOG.warning("The input tag is different from the tag in tar")

        # create container
        with docker_utils.docker_client() as docker:
            name = container.name
            LOG.debug('Creating container with image %(image)s name %(name)s',
                      {'image': image['image'], 'name': name})
            volmaps = requested_volumes.get(container.uuid, [])
            binds = self._get_binds(context, volmaps)
            kwargs = {
                'name': self.get_container_name(container),
                'command': container.command,
                'environment': container.environment,
                'working_dir': container.workdir,
                'labels': container.labels,
                'tty': container.tty,
                'stdin_open': container.interactive,
                'entrypoint': container.entrypoint,
            }

            host_config = {}
            host_config['privileged'] = container.privileged
            host_config['binds'] = binds
            kwargs['volumes'] = [b['bind'] for b in binds.values()]
            host_config['network_mode'] = 'container:%s' % capsule.container_id
            # TODO(hongbin): Uncomment this after docker-py add support for
            # container mode for pid namespace.
            # host_config['pid_mode'] = 'container:%s' % capsule.container_id
            host_config['ipc_mode'] = 'container:%s' % capsule.container_id
            if container.auto_remove:
                host_config['auto_remove'] = container.auto_remove
            if container.memory is not None:
                host_config['mem_limit'] = str(container.memory) + 'M'
            if container.cpu is not None:
                host_config['cpu_shares'] = int(1024 * container.cpu)
            if container.restart_policy:
                count = int(container.restart_policy['MaximumRetryCount'])
                name = container.restart_policy['Name']
                host_config['restart_policy'] = {'Name': name,
                                                 'MaximumRetryCount': count}

            if container.disk:
                disk_size = str(container.disk) + 'G'
                host_config['storage_opt'] = {'size': disk_size}
            # The time unit in docker of heath checking is us, and the unit
            # of interval and timeout is seconds.
            if container.healthcheck:
                healthcheck = {}
                healthcheck['test'] = container.healthcheck.get('test', '')
                interval = container.healthcheck.get('interval', 0)
                healthcheck['interval'] = interval * 10 ** 9
                healthcheck['retries'] = int(container.healthcheck.
                                             get('retries', 0))
                timeout = container.healthcheck.get('timeout', 0)
                healthcheck['timeout'] = timeout * 10 ** 9
                kwargs['healthcheck'] = healthcheck

            kwargs['host_config'] = docker.create_host_config(**host_config)
            if image['tag']:
                image_repo = image['repo'] + ":" + image['tag']
            else:
                image_repo = image['repo']
            response = docker.create_container(image_repo, **kwargs)
            container.container_id = response['Id']
            docker.start(container.container_id)

            response = docker.inspect_container(container.container_id)
            self._populate_container(container, response, force=True)
            container.save(context)

    def _wait_for_init_container(self, context, container, timeout=3600):
        def retry_if_result_is_false(result):
            return result is False

        def check_init_container_stopped():
            status = self.show(context, container).status
            if status == consts.STOPPED:
                return True
            elif status == consts.RUNNING:
                return False
            else:
                raise exception.ZunException(
                    _("Container has unexpected status: %s") % status)

        r = tenacity.Retrying(
            stop=tenacity.stop_after_delay(timeout),
            wait=tenacity.wait_exponential(),
            retry=tenacity.retry_if_result(retry_if_result_is_false))
        r.call(check_init_container_stopped)

    def delete_capsule(self, context, capsule, force):
        merged_containers = capsule.containers + capsule.init_containers
        for container in merged_containers:
            self._delete_container_in_capsule(context, capsule, container,
                                              force)
        with docker_utils.docker_client() as docker:
            try:
                docker.stop(capsule.container_id)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    pass
        self.delete(context, capsule, force)

    def _delete_container_in_capsule(self, context, capsule, container, force):
        if not container.container_id:
            return

        with docker_utils.docker_client() as docker:
            try:
                docker.stop(container.container_id)
                docker.remove_container(container.container_id,
                                        force=force)
            except errors.APIError as api_error:
                if is_not_found(api_error):
                    return
                if is_not_connected(api_error):
                    return
                raise



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zun.common.docker_image import digest
from zun.common.docker_image import reference
from zun.common import exception
from zun.tests import base


class TestReference(base.BaseTestCase):
    def test_reference(self):
        def create_test_case(input_, err=None, repository=None, hostname=None,
                             tag=None, digest=None, path=None):
            return {
                'input': input_,
                'err': err,
                'repository': repository,
                'hostname': hostname,
                'tag': tag,
                'digest': digest,
                'path': path,
            }

        test_cases = [
            create_test_case(input_='test_com', repository='test_com'),
            create_test_case(input_='test.com:tag', repository='test.com',
                             tag='tag'),
            create_test_case(input_='test.com:5000', repository='test.com',
                             tag='5000'),
            create_test_case(input_='test.com/repo:tag',
                             repository='test.com/repo', hostname='test.com',
                             tag='tag'),
            create_test_case(input_='test:5000/repo',
                             repository='test:5000/repo',
                             hostname='test:5000'),
            create_test_case(input_='test:5000/repo:tag',
                             repository='test:5000/repo', hostname='test:5000',
                             tag='tag'),
            create_test_case(input_='test:5000/repo@sha256:{}'.
                                    format('f' * 64),
                             repository='test:5000/repo',
                             hostname='test:5000',
                             digest='sha256:{}'.format('f' * 64)),
            create_test_case(input_='test:5000/repo:tag@sha256:{}'.
                                    format('f' * 64),
                             repository='test:5000/repo',
                             hostname='test:5000',
                             tag='tag',
                             digest='sha256:{}'.format('f' * 64)),
            create_test_case(input_='test:5000/repo',
                             repository='test:5000/repo',
                             hostname='test:5000'),
            create_test_case(input_='', err=exception.NameEmpty),
            create_test_case(input_=':justtag',
                             err=exception.ReferenceInvalidFormat),
            create_test_case(input_='@sha256:{}'.format('f' * 64),
                             err=exception.ReferenceInvalidFormat),
            create_test_case(input_='repo@sha256:{}'.format('f' * 34),
                             err=digest.DigestInvalidLength),
            create_test_case(input_='validname@invaliddigest:{}'.
                                    format('f' * 64),
                             err=digest.DigestUnsupported),
            create_test_case(input_='{}a:tag'.format('a/' * 128),
                             err=exception.NameTooLong),
            create_test_case(input_='{}a:tag-puts-this-over-max'.
                                    format('a/' * 127),
                             repository='{}a'.format('a/' * 127),
                             hostname='',
                             tag='tag-puts-this-over-max'),
            create_test_case(input_='aa/asdf$$^/aa',
                             err=exception.ReferenceInvalidFormat),
            create_test_case(input_='sub-dom1.foo.com/bar/baz/quux',
                             repository='sub-dom1.foo.com/bar/baz/quux',
                             hostname='sub-dom1.foo.com'),
            create_test_case(input_='sub-dom1.foo.com/bar/baz/quux:some-'
                                    'long-tag',
                             repository='sub-dom1.foo.com/bar/baz/quux',
                             hostname='sub-dom1.foo.com',
                             tag='some-long-tag'),
            create_test_case(input_='b.gcr.io/test.example.com/my-app:test.'
                                    'example.com',
                             repository='b.gcr.io/test.example.com/my-app',
                             hostname='b.gcr.io',
                             tag='test.example.com'),
            create_test_case(input_='xn--n3h.com/myimage:xn--n3h.com',
                             repository='xn--n3h.com/myimage',
                             hostname='xn--n3h.com',
                             tag='xn--n3h.com'),
            create_test_case(input_='xn--7o8h.com/myimage:xn--7o8h.com@'
                                    'sha512:{}'.format('f' * 128),
                             repository='xn--7o8h.com/myimage',
                             hostname='xn--7o8h.com',
                             tag='xn--7o8h.com',
                             digest='sha512:{}'.format('f' * 128)),
            create_test_case(input_='foo_bar.com:8080',
                             repository='foo_bar.com',
                             tag='8080'),
            create_test_case(input_='foo/foo_bar.com:8080',
                             repository='foo/foo_bar.com',
                             hostname='',
                             tag='8080'),
            create_test_case(input_='test.com/foo',
                             repository='test.com/foo',
                             hostname='test.com',
                             path='foo'),
            create_test_case(input_='test_com/foo',
                             repository='test_com/foo',
                             hostname='',
                             path='test_com/foo'),
            create_test_case(input_='test:8080/foo',
                             repository='test:8080/foo',
                             hostname='test:8080',
                             path='foo'),
            create_test_case(input_='test.com:8080/foo',
                             repository='test.com:8080/foo',
                             hostname='test.com:8080',
                             path='foo'),
            create_test_case(input_='test-com:8080/foo',
                             repository='test-com:8080/foo',
                             hostname='test-com:8080',
                             path='foo'),
            create_test_case(input_='xn--n3h.com:18080/foo',
                             repository='xn--n3h.com:18080/foo',
                             hostname='xn--n3h.com:18080',
                             path='foo'),
        ]

        for tc in test_cases:
            if tc['err']:
                self.assertRaises(tc['err'], reference.Reference.parse,
                                  tc['input'])
                continue

            try:
                r = reference.Reference.parse(tc['input'])
            except Exception as e:
                raise e
            else:
                if tc['repository'] is not None:
                    self.assertEqual(tc['repository'], r['name'])

                if tc['hostname'] is not None:
                    hostname, _ = r.split_hostname()
                    self.assertEqual(tc['hostname'], hostname)

                if tc['path'] is not None:
                    _, path = r.split_hostname()
                    self.assertEqual(tc['path'], path)

                if tc['tag'] is not None:
                    self.assertEqual(tc['tag'], r['tag'])

                if tc['digest'] is not None:
                    self.assertEqual(tc['digest'], r['digest'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestQuotaObject(base.DbTestCase):

    def setUp(self):
        super(TestQuotaObject, self).setUp()
        self.fake_quota = utils.get_test_quota()

    def test_get_quota(self):
        project_id = self.fake_quota['project_id']
        resource = self.fake_quota['resource']
        with mock.patch.object(self.dbapi, 'quota_get',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_quota
            quota = objects.Quota.get(self.context, project_id, resource)
            mock_get.assert_called_once_with(
                self.context, project_id, resource)
            self.assertEqual(self.context, quota._context)

    def test_get_all_quotas_by_project(self):
        project_id = self.fake_quota['project_id']
        with mock.patch.object(self.dbapi, 'quota_get_all_by_project',
                               autospec=True) as mock_get_all:
            mock_get_all.return_value = {
                'project_id': project_id,
                'resource_1': 10,
                'resource_2': 20
            }
            quotas_dict = objects.Quota.get_all(self.context, project_id)
            mock_get_all.assert_called_once_with(self.context, project_id)
            self.assertEqual(project_id, quotas_dict['project_id'])

    def test_create_quota(self):
        project_id = self.fake_quota['project_id']
        resource = self.fake_quota['resource']
        hard_limit = self.fake_quota['hard_limit']
        with mock.patch.object(self.dbapi, 'quota_create',
                               autospec=True) as mock_create:
            mock_create.return_value = self.fake_quota
            quota = objects.Quota(
                self.context, **utils.get_test_quota_value())
            quota.create(self.context)
            mock_create.assert_called_once_with(
                self.context, project_id, resource, hard_limit)

    def test_destroy_quota(self):
        project_id = self.fake_quota['project_id']
        resource = self.fake_quota['resource']
        with mock.patch.object(self.dbapi, 'quota_get',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_quota
            with mock.patch.object(self.dbapi, 'quota_destroy',
                                   autospec=True) as mock_destroy:
                quota = objects.Quota.get(self.context, project_id,
                                          resource)
                quota.destroy()
                mock_destroy.assert_called_once_with(
                    None, project_id, resource)
                self.assertEqual(self.context, quota._context)

    def test_update_quota(self):
        project_id = self.fake_quota['project_id']
        resource = self.fake_quota['resource']
        with mock.patch.object(self.dbapi, 'quota_get',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_quota
            with mock.patch.object(self.dbapi, 'quota_update',
                                   autospec=True) as mock_update:
                quota = objects.Quota.get(self.context, project_id,
                                          resource)
                quota.hard_limit = 100
                quota.update()

                mock_get.assert_called_once_with(self.context, project_id,
                                                 resource)
                mock_update.assert_called_once_with(
                    None, project_id, resource, 100)
                self.assertEqual(self.context, quota._context)
                self.assertEqual(100, quota.hard_limit)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add meta addresses to container

Revision ID: 4a0c4f7a4a33
Revises: 43e1088c3389
Create Date: 2016-11-20 12:18:44.086036

"""

# revision identifiers, used by Alembic.
revision = '4a0c4f7a4a33'
down_revision = '43e1088c3389'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from zun.db.sqlalchemy import models


def upgrade():
    op.add_column('container',
                  sa.Column('meta', models.JSONEncodedDict(),
                            nullable=True))
    op.add_column('container',
                  sa.Column('addresses', models.JSONEncodedDict(),
                            nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.api.controllers.v1.schemas import parameter_types
from zun.api.validation import validators
from zun.common import exception
from zun.tests import base


CONTAINER_CREATE = {
    'type': 'object',
    'properties': {
        'name': parameter_types.container_name,
        'image': parameter_types.image_name,
        'command': parameter_types.command_list,
        'cpu': parameter_types.cpu,
        'memory': parameter_types.memory,
        'workdir': parameter_types.workdir,
        'image_pull_policy': parameter_types.image_pull_policy,
        'labels': parameter_types.labels,
        'environment': parameter_types.environment,
        'restart_policy': parameter_types.restart_policy,
        'image_driver': parameter_types.image_driver,
        'security_groups': parameter_types.security_groups,
        'runtime': parameter_types.runtime,
        'auto_heal': parameter_types.boolean
    },
    'required': ['image'],
    'additionalProperties': False,
}

CAPSULE_CREATE = {
    'type': 'object',
    'properties': {
        'template': parameter_types.capsule_template
    },
    'required': ['template'],
    'additionalProperties': False
}


class TestSchemaValidations(base.BaseTestCase):
    def setUp(self):
        super(TestSchemaValidations, self).setUp()
        self.schema_validator = validators.SchemaValidator(CONTAINER_CREATE)

    def test_create_schema_with_all_valid_parameters(self):
        request_to_validate = {'name': 'test1', 'image': 'nginx',
                               'command': ["/bin/sh"],
                               'cpu': 1.0,
                               'memory': '5', 'workdir': '/workdir',
                               'image_pull_policy': 'never',
                               'labels': {'abc': 12, 'bcd': 'xyz'},
                               'environment': {'xyz': 'pqr', 'pqr': '2'},
                               'restart_policy': {'Name': 'no',
                                                  'MaximumRetryCount': '0'},
                               'image_driver': 'docker',
                               'security_groups': ['abc'],
                               'runtime': 'runc',
                               'auto_heal': False}
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_with_all_parameters_none(self):
        request_to_validate = {'name': None, 'image': 'nginx',
                               'command': None, 'cpu': None,
                               'memory': None, 'workdir': None,
                               'image_pull_policy': None,
                               'labels': None,
                               'environment': None,
                               'restart_policy': None,
                               'image_driver': None,
                               'security_groups': None,
                               'runtime': None,
                               'auto_heal': False
                               }
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_image_missing(self):
        request_to_validate = {'name': 'test1'}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'image' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_invalid_security_groups(self):
        invalid_security_groups = [[''], ['1' * 260], 'x']
        for value in invalid_security_groups:
            request_to_validate = {'security_groups': value, 'image': 'nginx'}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_invalid_name(self):
        invalid_names = ['a@', 'a', "", '*' * 265, " ", "     ", "a b", 'ab@']
        for value in invalid_names:
            request_to_validate = {'name': value, 'image': 'nginx'}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field 'name'"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_invalid_memory(self):
        invalid_memory = ['2KG', 2.2, '0', 0, '2', 2]
        for value in invalid_memory:
            request_to_validate = {'memory': value, 'image': 'nginx'}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field 'memory'"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_invalid_image_pull_policy(self):
        request_to_validate = {'name': 'test1', 'image': 'nginx',
                               'image_pull_policy': 'xyz'}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "Invalid input for field "
                                    "'image_pull_policy'"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_valid_memory(self):
        valid_memory = [4, 5, '4', '5', None]
        for value in valid_memory:
            request_to_validate = {'memory': value, 'image': 'nginx'}
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_cpu(self):
        valid_cpu = [4, 5, '4', '5', '1', '12.5', 1, 12.5, None]
        invalid_cpu = ['12a', 'abc', '0.a', 'a.90', "", "   ", " 0.9 "]
        for value in valid_cpu:
            request_to_validate = {'cpu': value, 'image': 'nginx'}
            self.schema_validator.validate(request_to_validate)
        for value in invalid_cpu:
            request_to_validate = {'cpu': value, 'image': 'nginx'}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field 'cpu'"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_restart_policy_name(self):
        valid_name = ['no', 'on-failure', 'unless-stopped', 'always']
        invalid_name = ['12a', 'abc', '5', 'dd', "", "   "]
        for value in valid_name:
            restart_policy = {'Name': value, 'MaximumRetryCount': '0'}
            request_to_validate = {'name': 'test1', 'image': 'nginx',
                                   'restart_policy': restart_policy}
            self.schema_validator.validate(request_to_validate)
        for value in invalid_name:
            restart_policy = {'Name': value, 'MaximumRetryCount': '0'}
            request_to_validate = {'name': 'test1', 'image': 'nginx',
                                   'restart_policy': restart_policy}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field 'Name'"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_restart_policy_maxiumretrycount(self):
        valid_retry = ['0', '1', 15, '100']
        invalid_retry = ['12a', 'abc', '-1', "", "   "]
        for value in valid_retry:
            restart_policy = {'Name': 'no', 'MaximumRetryCount': value}
            request_to_validate = {'name': 'test1', 'image': 'nginx',
                                   'restart_policy': restart_policy}
            self.schema_validator.validate(request_to_validate)
        for value in invalid_retry:
            restart_policy = {'Name': 'no', 'MaximumRetryCount': value}
            request_to_validate = {'name': 'test1', 'image': 'nginx',
                                   'restart_policy': restart_policy}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field "
                                        "'MaximumRetryCount'"):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_restart_policy(self):
        restart_policy = {'Name': 'no'}
        request_to_validate = {'name': 'test1', 'image': 'nginx',
                               'restart_policy': restart_policy}
        self.schema_validator.validate(request_to_validate)
        restart_policy = {'MaximumRetryCount': 5}
        request_to_validate = {'name': 'test1', 'image': 'nginx',
                               'restart_policy': restart_policy}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'Name' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_wrong_image_driver(self):
        request_to_validate = {'image_driver': 'xyz', 'image': 'nginx'}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "Invalid input for field"
                                    " 'image_driver'"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_wrong_environment(self):
        request_to_validate = {'image': 'nginx',
                               'environment': {'xyz': 'pqr', 'pqr': None}}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "Invalid input for field"
                                    " 'pqr'"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_wrong_runtime(self):
        request_to_validate = {'image': 'nginx',
                               'runtime': 123}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "Invalid input for field"
                                    " 'runtime'"):
            self.schema_validator.validate(request_to_validate)


class TestCapsuleSchemaValidations(base.BaseTestCase):
    def setUp(self):
        super(TestCapsuleSchemaValidations, self).setUp()
        self.schema_validator = validators.SchemaValidator(CAPSULE_CREATE)

    def test_create_schema_with_all_valid_parameters(self):
        request_to_validate = \
            {"template":
                {"kind": "capsule",
                 "capsuleVersion": "beta",
                 "metadata": {
                     "labels": {"app": "web"},
                     "name": "template"},

                 "spec": {
                     "restartPolicy": "Always",
                     "containers": [
                         {"workDir": "/root", "image": "ubuntu",
                          "volumeMounts": [{"readOnly": True,
                                            "mountPath": "/data1",
                                            "name": "volume1"}],
                          "command": ["/bin/bash"],
                          "env": {"ENV2": "/usr/bin"},
                          "imagePullPolicy": "ifnotpresent",
                          "ports": [{"containerPort": 80,
                                     "protocol": "TCP",
                                     "name": "nginx-port",
                                     "hostPort": 80}],
                          "resources": {"requests": {"cpu": 1,
                                                     "memory": 1024}}}],
                     "volumes": [
                         {"cinder": {"autoRemove": True, "size": 5},
                          "name": "volume1"},
                     ]}}}
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_kind_missing(self):
        request_to_validate = \
            {"template":
                {"capsuleVersion": "beta",
                 "metadata": {
                     "labels": {"app": "web"},
                     "name": "template"},
                 "spec": {"containers": [{"image": "test"}]}
                 }}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'kind' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_metadata_missing(self):
        request_to_validate = \
            {"template":
                {"capsuleVersion": "beta",
                 "kind": "capsule",
                 "spec": {"containers": [{"image": "test"}]}
                 }}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'metadata' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_spec_missing(self):
        request_to_validate = \
            {"template":
                {"capsuleVersion": "beta",
                 "kind": "capsule",
                 "metadata": {
                     "labels": {"app": "web"},
                     "name": "template"},
                 }}
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'spec' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_with_all_essential_params(self):
        request_to_validate = \
            {"template":
                {"kind": "capsule",
                 "capsuleVersion": "beta",
                 "metadata": {
                     "labels": {},
                     "name": "test-essential"},
                 "spec": {
                     "containers": [
                         {"image": "test"}]
                 }}}
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_capsule_restart_policy(self):
        valid_restart_policy = ["Always", "OnFailure", "Never"]
        invalid_restart_policy = ["always", "4", "onfailure", "never"]
        for restart_policy in valid_restart_policy:
            request_to_validate = \
                {"template":
                    {"capsuleVersion": "beta",
                     "kind": "capsule",
                     "metadata": {
                         "labels": {"app": "web"},
                         "name": "template"},
                     "spec": {"containers": [{"image": "test"}],
                              "restartPolicy": restart_policy}
                     }}
            self.schema_validator.validate(request_to_validate)
        for restart_policy in invalid_restart_policy:
            request_to_validate = \
                {"template":
                    {"capsuleVersion": "beta",
                     "kind": "capsule",
                     "metadata": {
                         "labels": {"app": "web"},
                         "name": "template"},
                     "spec": {"restartPolicy": restart_policy,
                              "containers": [{"image": "test"}]}
                     }}
            with self.assertRaisesRegex(exception.SchemaValidationError,
                                        "Invalid input for field "
                                        "'restartPolicy'."):
                self.schema_validator.validate(request_to_validate)

    def test_create_schema_capsule_existed_volume_mounts(self):
        request_to_validate = {
            "template": {
                "kind": "capsule",
                "metadata": {},
                "spec": {
                    "containers": [
                        {"image": "test",
                         "volumeMounts": [{
                             "name": "volume1",
                             "mountPath": "/data"}]
                         }],
                    "volumes": [
                        {"name": "volume1",
                         "cinder": {
                             "volumeID":
                                 "d2a28af0-e243-4525-adf9-2d091466e43d"}
                         }
                    ]
                }
            }
        }
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_capsule_new_volume_mounts(self):
        request_to_validate = {
            "template": {
                "kind": "capsule",
                "metadata": {},
                "spec": {
                    "containers": [
                        {"image": "test",
                         "volumeMounts": [{
                             "name": "volume1",
                             "mountPath": "/data"}]
                         }],
                    "volumes": [
                        {"name": "volume1",
                         "cinder": {
                             "size": 5,
                             "autoRemove": True}
                         }
                    ]
                }
            }
        }
        self.schema_validator.validate(request_to_validate)

    def test_create_schema_capsule_volume_no_cinder(self):
        request_to_validate = {
            "template": {
                "kind": "capsule",
                "metadata": {},
                "spec": {
                    "containers": [
                        {"image": "test"}],
                    "volumes": [
                        {"name": "volume1",
                         "no-cinder-driver": {
                             "size": 5,
                             "autoRemove": True}
                         }
                    ]
                }
            }
        }
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'cinder' is a required property"):
            self.schema_validator.validate(request_to_validate)

    def test_create_schema_capsule_volume_no_name(self):
        request_to_validate = {
            "template": {
                "kind": "capsule",
                "metadata": {},
                "spec": {
                    "containers": [
                        {"image": "test"}],
                    "volumes": [
                        {
                            "cinder": {
                                "size": 5,
                                "autoRemove": True}
                        }
                    ]
                }
            }
        }
        with self.assertRaisesRegex(exception.SchemaValidationError,
                                    "'name' is a required property"):
            self.schema_validator.validate(request_to_validate)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add pci device

Revision ID: 37bce72463e3
Revises: 10d65e285a59
Create Date: 2017-09-08 10:18:48.658980

"""

# revision identifiers, used by Alembic.
revision = '37bce72463e3'
down_revision = '10d65e285a59'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'pci_device',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36)),
        sa.Column('compute_node_uuid', sa.String(length=36), nullable=False),
        sa.Column('address', sa.String(length=12), nullable=False),
        sa.Column('vendor_id', sa.String(length=4), nullable=False),
        sa.Column('product_id', sa.String(length=4), nullable=False),
        sa.Column('dev_type', sa.String(length=8), nullable=False),
        sa.Column('dev_id', sa.String(255)),
        sa.Column('label', sa.String(255), nullable=False),
        sa.Column('status', sa.String(36), nullable=False),
        sa.Column('request_id', sa.String(36), nullable=False),
        sa.Column('extra_info', sa.Text()),
        sa.Column('container_uuid', sa.String(36)),
        sa.Column('numa_node', sa.Integer(), nullable=True),
        sa.Column('parent_addr', sa.String(12), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('compute_node_uuid', 'address',
                            name='uniq_pci_device0compute_node_uuid0address'),
        sa.Index('ix_pci_device_compute_node_uuid', 'compute_node_uuid'),
        sa.Index('ix_pci_device_container_uuid', 'container_uuid'),
        sa.Index('ix_pci_device_compute_node_uuid_parent_addr',
                 'compute_node_uuid', 'parent_addr'),
    )



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from zun.scheduler import filters

LOG = logging.getLogger(__name__)


class LabelFilter(filters.BaseHostFilter):
    """Filter the containers by label"""

    run_filter_once_per_request = True

    def host_passes(self, host_state, container, extra_spec):
        labels = {}
        for key, value in extra_spec['hints'].items():
            if key.startswith('label:'):
                newkey = key[6:]
                labels[newkey] = value

        if not labels:
            return True

        for key in labels:
            if not(key in host_state.labels and
                   host_state.labels.get(key) == labels.get(key)):
                LOG.debug("%(host_state)s does not have labels"
                          " %(key)s=%(value)s that container %(container)s"
                          " required.",
                          {'host_state': host_state,
                           'key': key,
                           'value': labels.get(key),
                           'container': container.name})
                return False

        return True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Create allocation table

Revision ID: e4d145e195f4
Revises: 09f196622a3f
Create Date: 2017-02-12 21:50:19.742699

"""

# revision identifiers, used by Alembic.
revision = 'e4d145e195f4'
down_revision = '09f196622a3f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    op.create_table(
        'allocation',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('resource_provider_id', sa.Integer(), nullable=False),
        sa.Column('consumer_id', sa.String(36), nullable=False),
        sa.Column('resource_class_id', sa.Integer(), nullable=False),
        sa.Column('used', sa.Integer(), nullable=False),
        sa.Column('is_nested', sa.Integer(), nullable=False),
        sa.Column('blob', zun.db.sqlalchemy.models.JSONEncodedList(),
                  nullable=True),
        sa.Index('allocation_resource_provider_class_used_idx',
                 'resource_provider_id', 'resource_class_id', 'used'),
        sa.Index('allocation_consumer_id_idx', 'consumer_id'),
        sa.Index('allocation_resource_class_id_idx', 'resource_class_id'),
        sa.PrimaryKeyConstraint('id'),
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Generic Loadable class support.

Meant to be used by such things as scheduler filters where we
want to load modules from certain directories and find certain types of
classes within those modules.  Note that this is quite different than
generic plugins and the pluginmanager code that exists elsewhere.

Usage:

Create a directory with an __init__.py with code such as:

class SomeLoadableClass(object):
    pass


class MyLoader(zun.loadables.BaseLoader)
    def __init__(self):
        super(MyLoader, self).__init__(SomeLoadableClass)

If you create modules in the same directory and subclass SomeLoadableClass
within them, MyLoader().get_all_classes() will return a list
of such classes.
"""

import inspect
import os
import sys

from oslo_utils import importutils

from zun.common import exception


class BaseLoader(object):
    def __init__(self, loadable_cls_type):
        mod = sys.modules[self.__class__.__module__]
        self.path = os.path.abspath(mod.__path__[0])
        self.package = mod.__package__
        self.loadable_cls_type = loadable_cls_type

    def _is_correct_class(self, obj):
        """Judge whether class type is correct

        Return whether an object is a class of the correct type and
        is not prefixed with an underscore.
        """
        return (inspect.isclass(obj) and
                (not obj.__name__.startswith('_')) and
                issubclass(obj, self.loadable_cls_type))

    def _get_classes_from_module(self, module_name):
        """Get the classes from a module that match the type we want."""
        classes = []
        module = importutils.import_module(module_name)
        for obj_name in dir(module):
            # Skip objects that are meant to be private.
            if obj_name.startswith('_'):
                continue
            itm = getattr(module, obj_name)
            if self._is_correct_class(itm):
                classes.append(itm)
        return classes

    def get_all_classes(self):
        """Get all classes

        Get the classes of the type we want from all modules found
        in the directory that defines this class.
        """
        classes = []
        for dirpath, dirnames, filenames in os.walk(self.path):
            relpath = os.path.relpath(dirpath, self.path)
            if relpath == '.':
                relpkg = ''
            else:
                relpkg = '.%s' % '.'.join(relpath.split(os.sep))
            for fname in filenames:
                root, ext = os.path.splitext(fname)
                if ext != '.py' or root == '__init__':
                    continue
                module_name = "%s%s.%s" % (self.package, relpkg, root)
                mod_classes = self._get_classes_from_module(module_name)
                classes.extend(mod_classes)
        return classes

    def get_matching_classes(self, loadable_class_names):
        """Get loadable classes from a list of names.

        Each name can be a full module path or the full path to a
        method that returns classes to use.  The latter behavior
        is useful to specify a method that returns a list of
        classes to use in a default case.
        """

        classes = []
        for cls_name in loadable_class_names:
            obj = importutils.import_class(cls_name)
            if self._is_correct_class(obj):
                classes.append(obj)
            elif inspect.isfunction(obj):
                # Get list of classes from a function
                for cls in obj():
                    classes.append(cls)
            else:
                error_str = 'Not a class of the correct type'
                raise exception.ClassNotFound(class_name=cls_name,
                                              exception=error_str)
        return classes



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add-entrypoint-to-container

Revision ID: f979327df44b
Revises: 74c97dca93d0
Create Date: 2020-03-30 03:19:01.080873

"""

# revision identifiers, used by Alembic.
revision = 'f979327df44b'
down_revision = '74c97dca93d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'entrypoint',
                zun.db.sqlalchemy.models.JSONEncodedList(),
                nullable=True))

    # ### end Alembic commands ###



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add resource provider table

Revision ID: 6fd4f7582eb0
Revises: 5458f8394206
Create Date: 2017-02-11 10:01:31.713279

"""

# revision identifiers, used by Alembic.
revision = '6fd4f7582eb0'
down_revision = '5458f8394206'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'resource_provider',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('root_provider', sa.String(length=36), nullable=False),
        sa.Column('parent_provider', sa.String(length=36), nullable=True),
        sa.Column('can_host', sa.Integer(), default=0),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_resource_provider0uuid')
    )



# Copyright 2016 IBM, Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zun.common import exception
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class DbZunServiceTestCase(base.DbTestCase):

    def setUp(self):
        super(DbZunServiceTestCase, self).setUp()

    def test_create_zun_service(self):
        utils.create_test_zun_service()

    def test_create_zun_service_failure_for_dup(self):
        utils.create_test_zun_service()
        self.assertRaises(exception.ZunServiceAlreadyExists,
                          utils.create_test_zun_service)

    def test_get_zun_service(self):
        ms = utils.create_test_zun_service()
        res = self.dbapi.get_zun_service(
            ms['host'], ms['binary'])
        self.assertEqual(ms.id, res.id)

    def test_get_zun_service_failure(self):
        utils.create_test_zun_service()
        res = self.dbapi.get_zun_service(
            'fakehost1', 'fake-bin1')
        self.assertIsNone(res)

    def test_update_zun_service(self):
        ms = utils.create_test_zun_service()
        d2 = True
        update = {'disabled': d2}
        ms1 = self.dbapi.update_zun_service(ms['host'], ms['binary'], update)
        self.assertEqual(ms['id'], ms1['id'])
        self.assertEqual(d2, ms1['disabled'])
        res = self.dbapi.get_zun_service(
            'fakehost', 'fake-bin')
        self.assertEqual(ms1['id'], res['id'])
        self.assertEqual(d2, res['disabled'])

    def test_update_zun_service_failure(self):
        fake_update = {'fake_field': 'fake_value'}
        self.assertRaises(exception.ZunServiceNotFound,
                          self.dbapi.update_zun_service,
                          'fakehost1', 'fake-bin1', fake_update)

    def test_destroy_zun_service(self):
        ms = utils.create_test_zun_service()
        res = self.dbapi.get_zun_service(
            'fakehost', 'fake-bin')
        self.assertEqual(res['id'], ms['id'])
        self.dbapi.destroy_zun_service(ms['host'], ms['binary'])
        res = self.dbapi.get_zun_service(
            'fakehost', 'fake-bin')
        self.assertIsNone(res)

    def test_destroy_zun_service_failure(self):
        self.assertRaises(exception.ZunServiceNotFound,
                          self.dbapi.destroy_zun_service,
                          'fakehostsssss', 'fakessss-bin1')

    def test_list_zun_services(self):
        fake_ms_params = {
            'report_count': 1010,
            'host': 'FakeHost',
            'binary': 'FakeBin',
            'disabled': False,
            'disabled_reason': 'FakeReason'
        }
        utils.create_test_zun_service(**fake_ms_params)
        res = self.dbapi.list_zun_services()
        self.assertEqual(1, len(res))
        res = res[0]
        for k, v in fake_ms_params.items():
            self.assertEqual(res[k], v)

        fake_ms_params['binary'] = 'FakeBin1'
        fake_ms_params['disabled'] = True
        utils.create_test_zun_service(**fake_ms_params)
        res = self.dbapi.list_zun_services(filters={'disabled': True})
        self.assertEqual(1, len(res))
        res = res[0]
        for k, v in fake_ms_params.items():
            self.assertEqual(res[k], v)

    def test_list_zun_services_by_binary(self):
        fake_ms_params = {
            'report_count': 1010,
            'host': 'FakeHost',
            'binary': 'FakeBin',
            'disabled': False,
            'disabled_reason': 'FakeReason'
        }
        utils.create_test_zun_service(**fake_ms_params)
        res = self.dbapi.list_zun_services_by_binary(
            binary=fake_ms_params['binary'])
        self.assertEqual(1, len(res))
        res = res[0]
        for k, v in fake_ms_params.items():
            self.assertEqual(res[k], v)

        res = self.dbapi.list_zun_services_by_binary(binary='none')
        self.assertEqual(0, len(res))



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from zun.api import servicegroup
from zun.common import exception
from zun.common import profiler
from zun.common import rpc_service
import zun.conf
from zun import objects


def check_container_host(func):
    """Verify the state of container host"""
    @functools.wraps(func)
    def wrap(self, context, container, *args, **kwargs):
        services = objects.ZunService.list_by_binary(context, 'zun-compute')
        api_servicegroup = servicegroup.ServiceGroup()
        up_hosts = [service.host for service in services
                    if api_servicegroup.service_is_up(service)]
        if container.host is not None and container.host not in up_hosts:
            raise exception.ContainerHostNotUp(container=container.uuid,
                                               host=container.host)
        return func(self, context, container, *args, **kwargs)
    return wrap


@profiler.trace_cls("rpc")
class API(rpc_service.API):
    """Client side of the container compute rpc API.

    API version history:

        * 1.0 - Initial version.
        * 1.1 - Add image endpoints.
    """

    def __init__(self, context=None, topic=None):
        if topic is None:
            zun.conf.CONF.import_opt(
                'topic', 'zun.conf.compute', group='compute')

        super(API, self).__init__(
            context, topic=zun.conf.CONF.compute.topic)

    def container_create(self, context, host, container, limits,
                         requested_networks, requested_volumes, run,
                         pci_requests):
        self._cast(host, 'container_create', limits=limits,
                   requested_networks=requested_networks,
                   requested_volumes=requested_volumes,
                   container=container,
                   run=run,
                   pci_requests=pci_requests)

    @check_container_host
    def container_delete(self, context, container, force=False):
        return self._cast(container.host, 'container_delete',
                          container=container, force=force)

    @check_container_host
    def container_show(self, context, container):
        return self._call(container.host, 'container_show',
                          container=container)

    def container_rebuild(self, context, container, run):
        self._cast(container.host, 'container_rebuild', container=container,
                   run=run)

    def container_reboot(self, context, container, timeout):
        self._cast(container.host, 'container_reboot', container=container,
                   timeout=timeout)

    def container_stop(self, context, container, timeout):
        self._cast(container.host, 'container_stop', container=container,
                   timeout=timeout)

    def container_start(self, context, container):
        self._cast(container.host, 'container_start', container=container)

    def container_pause(self, context, container):
        self._cast(container.host, 'container_pause', container=container)

    def container_unpause(self, context, container):
        self._cast(container.host, 'container_unpause', container=container)

    @check_container_host
    def container_logs(self, context, container, stdout, stderr,
                       timestamps, tail, since):
        return self._call(container.host, 'container_logs',
                          container=container, stdout=stdout, stderr=stderr,
                          timestamps=timestamps, tail=tail, since=since)

    @check_container_host
    def container_exec(self, context, container, command, run, interactive):
        return self._call(container.host, 'container_exec',
                          container=container, command=command, run=run,
                          interactive=interactive)

    @check_container_host
    def container_exec_resize(self, context, container, exec_id, height,
                              width):
        return self._call(container.host, 'container_exec_resize',
                          exec_id=exec_id, height=height, width=width)

    def container_kill(self, context, container, signal):
        self._cast(container.host, 'container_kill', container=container,
                   signal=signal)

    @check_container_host
    def container_update(self, context, container, patch):
        return self._call(container.host, 'container_update',
                          container=container, patch=patch)

    def resize_container(self, context, container, patch):
        self._cast(container.host, 'resize_container',
                   container=container, patch=patch)

    @check_container_host
    def container_attach(self, context, container):
        return self._call(container.host, 'container_attach',
                          container=container)

    @check_container_host
    def container_resize(self, context, container, height, width):
        return self._call(container.host, 'container_resize',
                          container=container, height=height, width=width)

    @check_container_host
    def container_top(self, context, container, ps_args):
        return self._call(container.host, 'container_top',
                          container=container, ps_args=ps_args)

    @check_container_host
    def container_get_archive(self, context, container, path, encode_data):
        return self._call(container.host, 'container_get_archive',
                          container=container, path=path,
                          encode_data=encode_data)

    @check_container_host
    def container_put_archive(self, context, container, path, data,
                              decode_data):
        return self._call(container.host, 'container_put_archive',
                          container=container, path=path, data=data,
                          decode_data=decode_data)

    @check_container_host
    def container_stats(self, context, container):
        return self._call(container.host, 'container_stats',
                          container=container)

    @check_container_host
    def container_commit(self, context, container, repository, tag):
        return self._call(container.host, 'container_commit',
                          container=container, repository=repository, tag=tag)

    def add_security_group(self, context, container, security_group):
        return self._cast(container.host, 'add_security_group',
                          container=container, security_group=security_group)

    def remove_security_group(self, context, container, security_group):
        return self._cast(container.host, 'remove_security_group',
                          container=container, security_group=security_group)

    def image_delete(self, context, image, host):
        self._cast(host, 'image_delete', image=image)

    def image_pull(self, context, image, host):
        self._cast(host, 'image_pull', image=image)

    def image_search(self, context, image, image_driver, exact_match,
                     registry, host=None):
        return self._call(host, 'image_search', image=image,
                          image_driver_name=image_driver,
                          exact_match=exact_match,
                          registry=registry)

    def capsule_create(self, context, host, capsule,
                       requested_networks, requested_volumes, limits):
        self._cast(host, 'capsule_create',
                   capsule=capsule,
                   requested_networks=requested_networks,
                   requested_volumes=requested_volumes,
                   limits=limits)

    def capsule_delete(self, context, capsule):
        return self._call(capsule.host, 'capsule_delete',
                          capsule=capsule)

    def network_detach(self, context, container, network):
        self._cast(container.host, 'network_detach',
                   container=container, network=network)

    def network_attach(self, context, container, requested_network):
        self._cast(container.host, 'network_attach',
                   container=container,
                   requested_network=requested_network)

    def network_create(self, context, neutron_net_id):
        host = None
        return self._call(host, 'network_create',
                          neutron_net_id=neutron_net_id)

    def network_delete(self, context, network):
        host = None
        return self._call(host, 'network_delete',
                          network=network)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add host to ZunNetwork

Revision ID: 3f2b36231bee
Revises: f979327df44b
Create Date: 2023-12-18 10:47:27.164812

"""

# revision identifiers, used by Alembic.
revision = '3f2b36231bee'
down_revision = 'f979327df44b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('network',
                  sa.Column('host', sa.String(length=255), nullable=True))
    op.drop_constraint(constraint_name='uniq_network0neutron_net_id',
                       table_name='network', type_='unique')
    op.create_unique_constraint(
        constraint_name='uniq_network0neutron_net_id_host',
        table_name='network', columns=['neutron_net_id', 'host'])



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""rename volume_id to cinder_volume_id in volume_mapping

Revision ID: 35cb52c5553f
Revises: 54bcb75afb32
Create Date: 2018-09-22 20:20:02.072979

"""

# revision identifiers, used by Alembic.
revision = '35cb52c5553f'
down_revision = '54bcb75afb32'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column(
        'volume_mapping', 'volume_id',
        new_column_name='cinder_volume_id',
        type_=sa.String(36)
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""create_table_zun_service

Revision ID: a9a92eebd9a8
Revises:
Create Date: 2016-05-27 10:53:10.367610

"""

# revision identifiers, used by Alembic.
revision = 'a9a92eebd9a8'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'zun_service',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('report_count', sa.Integer(), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=True),
        sa.Column('binary', sa.String(length=255), nullable=True),
        sa.Column('disabled', sa.Boolean(), nullable=True),
        sa.Column('disabled_reason', sa.String(length=255), nullable=True),
        sa.Column('last_seen_up', sa.DateTime(), nullable=True),
        sa.Column('forced_down', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('host', 'binary',
                            name='uniq_zun_service0host0binary')
    )



# Copyright (c) 2017 Intel, Inc.
# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import builtins
import glob
import os
from unittest import mock

import fixtures

from zun.common import exception
from zun.pci import utils
from zun.tests import base


class PciDeviceMatchTestCase(base.TestCase):
    def setUp(self):
        super(PciDeviceMatchTestCase, self).setUp()
        self.fake_pci_1 = {'vendor_id': 'v1',
                           'device_id': 'd1',
                           'capabilities_network': ['cap1', 'cap2', 'cap3']}

    def test_single_spec_match(self):
        self.assertTrue(utils.pci_device_prop_match(
            self.fake_pci_1, [{'vendor_id': 'v1', 'device_id': 'd1'}]))

    def test_multiple_spec_match(self):
        self.assertTrue(utils.pci_device_prop_match(
            self.fake_pci_1,
            [{'vendor_id': 'v1', 'device_id': 'd1'},
             {'vendor_id': 'v3', 'device_id': 'd3'}]))

    def test_spec_dismatch(self):
        self.assertFalse(utils.pci_device_prop_match(
            self.fake_pci_1,
            [{'vendor_id': 'v4', 'device_id': 'd4'},
             {'vendor_id': 'v3', 'device_id': 'd3'}]))

    def test_spec_extra_key(self):
        self.assertFalse(utils.pci_device_prop_match(
            self.fake_pci_1,
            [{'vendor_id': 'v1', 'device_id': 'd1', 'wrong_key': 'k1'}]))

    def test_spec_list(self):
        self.assertTrue(utils.pci_device_prop_match(
            self.fake_pci_1, [{'vendor_id': 'v1', 'device_id': 'd1',
                               'capabilities_network': ['cap1', 'cap2',
                                                        'cap3']}]))
        self.assertTrue(utils.pci_device_prop_match(
            self.fake_pci_1, [{'vendor_id': 'v1', 'device_id': 'd1',
                               'capabilities_network': ['cap3', 'cap1']}]))

    def test_spec_list_no_matching(self):
        self.assertFalse(utils.pci_device_prop_match(
            self.fake_pci_1, [{'vendor_id': 'v1', 'device_id': 'd1',
                               'capabilities_network': ['cap1', 'cap33']}]))

    def test_spec_list_wrong_type(self):
        self.assertFalse(utils.pci_device_prop_match(
            self.fake_pci_1, [{'vendor_id': 'v1', 'device_id': ['d1']}]))


class PciDeviceAddressParserTestCase(base.TestCase):
    def test_parse_address(self):
        self.parse_result = utils.parse_address("0000:04:12.6")
        self.assertEqual(self.parse_result, ('0000', '04', '12', '6'))

    def test_parse_address_wrong(self):
        self.assertRaises(exception.PciDeviceWrongAddressFormat,
                          utils.parse_address, "0000:04.12:6")

    def test_parse_address_invalid_character(self):
        self.assertRaises(exception.PciDeviceWrongAddressFormat,
                          utils.parse_address, "0000:h4.12:6")


class GetFunctionByIfnameTestCase(base.TestCase):

    @mock.patch('os.path.isdir', return_value=True)
    @mock.patch.object(os, 'readlink')
    def test_virtual_function(self, mock_readlink, *args):
        mock_readlink.return_value = '../../../0000.00.00.1'
        with mock.patch.object(builtins, 'open', side_effect=IOError()):
            address, physical_function = utils.get_function_by_ifname('eth0')
            self.assertEqual(address, '0000.00.00.1')
            self.assertFalse(physical_function)

    @mock.patch('os.path.isdir', return_value=True)
    @mock.patch.object(os, 'readlink')
    def test_physical_function(self, mock_readlink, *args):
        ifname = 'eth0'
        totalvf_path = "/sys/class/net/%s/device/%s" % (ifname,
                                                        utils._SRIOV_TOTALVFS)
        mock_readlink.return_value = '../../../0000:00:00.1'
        with mock.patch.object(builtins, 'open',
                               mock.mock_open(read_data='4')) as mock_open:
            address, physical_function = utils.get_function_by_ifname('eth0')
            self.assertEqual(address, '0000:00:00.1')
            self.assertTrue(physical_function)
            mock_open.assert_called_once_with(totalvf_path)

    @mock.patch('os.path.isdir', return_value=False)
    def test_exception(self, *args):
        address, physical_function = utils.get_function_by_ifname('lo')
        self.assertIsNone(address)
        self.assertFalse(physical_function)


class IsPhysicalFunctionTestCase(base.TestCase):

    def setUp(self):
        super(IsPhysicalFunctionTestCase, self).setUp()
        self.pci_args = utils.get_pci_address_fields('0000:00:00.1')

    @mock.patch('os.path.isdir', return_value=True)
    def test_virtual_function(self, *args):
        with mock.patch.object(builtins, 'open', side_effect=IOError()):
            self.assertFalse(utils.is_physical_function(*self.pci_args))

    @mock.patch('os.path.isdir', return_value=True)
    def test_physical_function(self, *args):
        with mock.patch.object(builtins, 'open',
                               mock.mock_open(read_data='4')):
            self.assertTrue(utils.is_physical_function(*self.pci_args))

    @mock.patch('os.path.isdir', return_value=False)
    def test_exception(self, *args):
        self.assertFalse(utils.is_physical_function(*self.pci_args))


class GetIfnameByPciAddressTestCase(base.TestCase):

    def setUp(self):
        super(GetIfnameByPciAddressTestCase, self).setUp()
        self.pci_address = '0000:00:00.1'

    @mock.patch.object(os, 'listdir')
    def test_physical_function_inferface_name(self, mock_listdir):
        mock_listdir.return_value = ['foo', 'bar']
        ifname = utils.get_ifname_by_pci_address(
            self.pci_address, pf_interface=True)
        self.assertEqual(ifname, 'bar')

    @mock.patch.object(os, 'listdir')
    def test_virtual_function_inferface_name(self, mock_listdir):
        mock_listdir.return_value = ['foo', 'bar']
        ifname = utils.get_ifname_by_pci_address(
            self.pci_address, pf_interface=False)
        self.assertEqual(ifname, 'bar')

    @mock.patch.object(os, 'listdir')
    def test_exception(self, mock_listdir):
        mock_listdir.side_effect = OSError('No such file or directory')
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            utils.get_ifname_by_pci_address,
            self.pci_address
        )


class GetMacByPciAddressTestCase(base.TestCase):
    def setUp(self):
        super(GetMacByPciAddressTestCase, self).setUp()
        self.pci_address = '0000:07:00.1'
        self.if_name = 'enp7s0f1'
        self.tmpdir = self.useFixture(fixtures.TempDir())
        self.fake_file = os.path.join(self.tmpdir.path, "address")
        with open(self.fake_file, "w") as f:
            f.write("a0:36:9f:72:00:00\n")

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(os.path, 'join')
    def test_get_mac(self, mock_join, mock_listdir):
        mock_listdir.return_value = [self.if_name]
        mock_join.return_value = self.fake_file
        mac = utils.get_mac_by_pci_address(self.pci_address)
        mock_join.assert_called_once_with(
            "/sys/bus/pci/devices/%s/net" % self.pci_address, self.if_name,
            "address")
        self.assertEqual("a0:36:9f:72:00:00", mac)

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(os.path, 'join')
    def test_get_mac_fails(self, mock_join, mock_listdir):
        os.unlink(self.fake_file)
        mock_listdir.return_value = [self.if_name]
        mock_join.return_value = self.fake_file
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            utils.get_mac_by_pci_address, self.pci_address)

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(os.path, 'join')
    def test_get_mac_fails_empty(self, mock_join, mock_listdir):
        with open(self.fake_file, "w") as f:
            f.truncate(0)
        mock_listdir.return_value = [self.if_name]
        mock_join.return_value = self.fake_file
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            utils.get_mac_by_pci_address, self.pci_address)

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(os.path, 'join')
    def test_get_physical_function_mac(self, mock_join, mock_listdir):
        mock_listdir.return_value = [self.if_name]
        mock_join.return_value = self.fake_file
        mac = utils.get_mac_by_pci_address(self.pci_address, pf_interface=True)
        mock_join.assert_called_once_with(
            "/sys/bus/pci/devices/%s/physfn/net" % self.pci_address,
            self.if_name, "address")
        self.assertEqual("a0:36:9f:72:00:00", mac)


class GetVfNumByPciAddressTestCase(base.TestCase):

    def setUp(self):
        super(GetVfNumByPciAddressTestCase, self).setUp()
        self.pci_address = '0000:00:00.1'
        self.paths = [
            '/sys/bus/pci/devices/0000:00:00.1/physfn/virtfn3',
        ]

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_vf_number_found(self, mock_iglob, mock_readlink):
        mock_iglob.return_value = self.paths
        mock_readlink.return_value = '../../0000:00:00.1'
        vf_num = utils.get_vf_num_by_pci_address(self.pci_address)
        self.assertEqual(vf_num, '3')

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_vf_number_not_found(self, mock_iglob, mock_readlink):
        mock_iglob.return_value = self.paths
        mock_readlink.return_value = '../../0000:00:00.2'
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            utils.get_vf_num_by_pci_address,
            self.pci_address
        )

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_exception(self, mock_iglob, mock_readlink):
        mock_iglob.return_value = self.paths
        mock_readlink.side_effect = OSError('No such file or directory')
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            utils.get_vf_num_by_pci_address,
            self.pci_address
        )


class GetNetNameByVfPciAddressTestCase(base.TestCase):

    def setUp(self):
        super(GetNetNameByVfPciAddressTestCase, self).setUp()
        self._get_mac = mock.patch.object(utils, 'get_mac_by_pci_address')
        self.mock_get_mac = self._get_mac.start()
        self._get_ifname = mock.patch.object(
            utils, 'get_ifname_by_pci_address')
        self.mock_get_ifname = self._get_ifname.start()
        self.addCleanup(self._get_mac.stop)
        self.addCleanup(self._get_ifname.stop)

        self.mac = 'ca:fe:ca:fe:ca:fe'
        self.if_name = 'enp7s0f0'
        self.pci_address = '0000:07:02.1'

    def test_correct_behaviour(self):
        ref_net_name = ('net_enp7s0f0_ca_fe_ca_fe_ca_fe', 'enp7s0f0')
        self.mock_get_mac.return_value = self.mac
        self.mock_get_ifname.return_value = self.if_name
        net_name = utils.get_net_name_by_vf_pci_address(self.pci_address)
        self.assertEqual(ref_net_name, net_name)
        self.mock_get_mac.assert_called_once_with(self.pci_address)
        self.mock_get_ifname.assert_called_once_with(self.pci_address)

    def test_wrong_mac(self):
        self.mock_get_mac.side_effect = (
            exception.PciDeviceNotFoundById(self.pci_address))
        net_name = utils.get_net_name_by_vf_pci_address(self.pci_address)
        self.assertIsNone(net_name)
        self.mock_get_mac.assert_called_once_with(self.pci_address)
        self.mock_get_ifname.assert_not_called()

    def test_wrong_ifname(self):
        self.mock_get_mac.return_value = self.mac
        self.mock_get_ifname.side_effect = (
            exception.PciDeviceNotFoundById(self.pci_address))
        net_name = utils.get_net_name_by_vf_pci_address(self.pci_address)
        self.assertIsNone(net_name)
        self.mock_get_mac.assert_called_once_with(self.pci_address)
        self.mock_get_ifname.assert_called_once_with(self.pci_address)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbExecInstanceTestCase(base.DbTestCase):

    def setUp(self):
        super(DbExecInstanceTestCase, self).setUp()

    def test_create_exec_instance(self):
        utils.create_test_exec_instance(context=self.context)

    def test_create_exec_instance_already_exists(self):
        utils.create_test_exec_instance(context=self.context,
                                        container_id=1, exec_id='test-id')
        with self.assertRaisesRegex(
                exception.ExecInstanceAlreadyExists,
                'An exec instance with exec_id test-id .*'):
            utils.create_test_exec_instance(
                context=self.context, container_id=1, exec_id='test-id')

    def test_list_exec_instances(self):
        exec_ids = []
        for i in range(1, 6):
            exec_inst = utils.create_test_exec_instance(
                id=i,
                context=self.context,
                container_id=1,
                exec_id=uuidutils.generate_uuid())
            exec_ids.append(str(exec_inst['exec_id']))
        res = dbapi.list_exec_instances(self.context)
        res_exec_ids = [r.exec_id for r in res]
        self.assertEqual(sorted(exec_ids), sorted(res_exec_ids))

    def test_list_exec_instances_sorted(self):
        exec_ids = []
        for i in range(5):
            exec_inst = utils.create_test_exec_instance(
                id=i,
                context=self.context,
                container_id=1,
                exec_id=uuidutils.generate_uuid())
            exec_ids.append(str(exec_inst['exec_id']))
        res = dbapi.list_exec_instances(self.context, sort_key='exec_id')
        res_exec_ids = [r.exec_id for r in res]
        self.assertEqual(sorted(exec_ids), res_exec_ids)

    def test_list_exec_instances_with_filters(self):
        exec_inst1 = utils.create_test_exec_instance(
            id=1,
            context=self.context,
            container_id=1,
            exec_id='exec-one')
        exec_inst2 = utils.create_test_exec_instance(
            id=2,
            context=self.context,
            container_id=2,
            exec_id='exec-two')

        res = dbapi.list_exec_instances(
            self.context, filters={'container_id': 1})
        self.assertEqual([exec_inst1.id], [r.id for r in res])

        res = dbapi.list_exec_instances(
            self.context, filters={'container_id': 2})
        self.assertEqual([exec_inst2.id], [r.id for r in res])

        res = dbapi.list_exec_instances(
            self.context, filters={'container_id': 777})
        self.assertEqual([], [r.id for r in res])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinderclient import client as cinderclient
from glanceclient import client as glanceclient
from neutronclient.v2_0 import client as neutronclient

from zun.common import clients
import zun.conf
from zun.tests import base


class ClientsTest(base.BaseTestCase):

    def setUp(self):
        super(ClientsTest, self).setUp()

        zun.conf.CONF.set_override('www_authenticate_uri',
                                   'http://server.test:5000/v2.0',
                                   group='keystone_authtoken')
        zun.conf.CONF.import_opt('api_version', 'zun.conf.glance_client',
                                 group='glance_client')

    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def test_url_for(self, mock_keystone):
        obj = clients.OpenStackClients(None)
        obj.url_for(service_type='fake_service', interface='fake_endpoint')

        mock_endpoint = mock_keystone.return_value.session.get_endpoint
        mock_endpoint.assert_called_once_with(service_type='fake_service',
                                              interface='fake_endpoint')

    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def test_zun_url(self, mock_keystone):
        fake_region = 'fake_region'
        fake_endpoint = 'fake_endpoint'
        zun.conf.CONF.set_override('region_name', fake_region,
                                   group='zun_client')
        zun.conf.CONF.set_override('endpoint_type', fake_endpoint,
                                   group='zun_client')
        obj = clients.OpenStackClients(None)
        obj.zun_url()

        mock_endpoint = mock_keystone.return_value.session.get_endpoint
        mock_endpoint.assert_called_once_with(region_name=fake_region,
                                              service_type='container',
                                              interface=fake_endpoint)

    @mock.patch.object(glanceclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def _test_clients_glance(self, expected_region_name, mock_keystone,
                             mock_call):
        mock_session = mock_keystone.return_value.session
        mock_session.get_endpoint.return_value = 'glance_url'
        fake_region = 'fake_region'
        fake_endpoint = 'fake_endpoint'
        zun.conf.CONF.set_override('region_name', fake_region,
                                   group='glance_client')
        zun.conf.CONF.set_override('endpoint_type', fake_endpoint,
                                   group='glance_client')
        con = mock.MagicMock()
        obj = clients.OpenStackClients(con)
        obj._glance = None
        obj.glance()

        mock_session.get_endpoint.assert_called_once_with(
            region_name=fake_region, service_type='image',
            interface=fake_endpoint)
        mock_call.assert_called_once_with(
            zun.conf.CONF.glance_client.api_version,
            endpoint='glance_url',
            session=mock_session)

    def test_clients_glance(self):
        self._test_clients_glance(None)

    def test_clients_glance_region(self):
        zun.conf.CONF.set_override('region_name',
                                   'myregion', group='glance_client')
        self._test_clients_glance('myregion')

    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def test_clients_glance_cached(self, mock_keystone):
        mock_keystone.return_value.session.get_endpoint.return_value = \
            'glance_url'
        con = mock.MagicMock()
        obj = clients.OpenStackClients(con)
        obj._glance = None
        glance = obj.glance()
        glance_cached = obj.glance()
        self.assertEqual(glance, glance_cached)

    @mock.patch.object(neutronclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def test_clients_neturon(self, mock_keystone, mock_call):
        ca_file = '/testpath'
        insecure = True
        zun.conf.CONF.set_override('ca_file', ca_file,
                                   group='neutron_client')
        zun.conf.CONF.set_override('insecure', insecure,
                                   group='neutron_client')
        con = mock.MagicMock()
        con.auth_token = "3bcc3d3a03f44e3d8377f9247b0ad155"
        con.auth_url = "keystone_url"
        obj = clients.OpenStackClients(con)
        obj._neutron = None
        obj.neutron()
        mock_call.assert_called_once()

    @mock.patch.object(cinderclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'keystone')
    def test_clients_cinder(self, mock_keystone, mock_call):
        ca_file = '/testpath'
        insecure = True
        endpoint_type = 'internalURL'
        zun.conf.CONF.set_override('ca_file', ca_file,
                                   group='cinder_client')
        zun.conf.CONF.set_override('insecure', insecure,
                                   group='cinder_client')
        zun.conf.CONF.set_override('endpoint_type', endpoint_type,
                                   group='cinder_client')
        con = mock.MagicMock()
        con.auth_token = "3bcc3d3a03f44e3d8377f9247b0ad155"
        con.auth_url = "keystone_url"
        obj = clients.OpenStackClients(con)
        obj._cinder = None
        obj.cinder()
        mock_call.assert_called_once()



# Copyright 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils as json

from zun.common import context
from zun.common import rpc
from zun.tests import base


class TestRpc(base.TestCase):

    def test_add_extra_exmods(self):
        rpc.EXTRA_EXMODS = []

        rpc.add_extra_exmods('foo', 'bar')

        self.assertEqual(['foo', 'bar'], rpc.EXTRA_EXMODS)

    def test_clear_extra_exmods(self):
        rpc.EXTRA_EXMODS = ['foo', 'bar']

        rpc.clear_extra_exmods()

        self.assertEqual(0, len(rpc.EXTRA_EXMODS))

    def test_serialize_entity(self):
        with mock.patch.object(json, 'to_primitive') as mock_prim:
            rpc.JsonPayloadSerializer.serialize_entity('context', 'entity')

        mock_prim.assert_called_once_with('entity', convert_instances=True)


class TestRequestContextSerializer(base.TestCase):
    def setUp(self):
        super(TestRequestContextSerializer, self).setUp()
        self.mock_base = mock.Mock()
        self.ser = rpc.RequestContextSerializer(self.mock_base)
        self.ser_null = rpc.RequestContextSerializer(None)

    def test_serialize_entity(self):
        self.mock_base.serialize_entity.return_value = 'foo'

        ser_ent = self.ser.serialize_entity('context', 'entity')

        self.mock_base.serialize_entity.assert_called_once_with('context',
                                                                'entity')
        self.assertEqual('foo', ser_ent)

    def test_serialize_entity_null_base(self):
        ser_ent = self.ser_null.serialize_entity('context', 'entity')

        self.assertEqual('entity', ser_ent)

    def test_deserialize_entity(self):
        self.mock_base.deserialize_entity.return_value = 'foo'

        deser_ent = self.ser.deserialize_entity('context', 'entity')

        self.mock_base.deserialize_entity.assert_called_once_with('context',
                                                                  'entity')
        self.assertEqual('foo', deser_ent)

    def test_deserialize_entity_null_base(self):
        deser_ent = self.ser_null.deserialize_entity('context', 'entity')

        self.assertEqual('entity', deser_ent)

    def test_serialize_context(self):
        context = mock.Mock()

        self.ser.serialize_context(context)

        context.to_dict.assert_called_once_with()

    @mock.patch.object(context, 'RequestContext')
    def test_deserialize_context(self, mock_req):
        self.ser.deserialize_context('context')

        mock_req.from_dict.assert_called_once_with('context')


class TestProfilerRequestContextSerializer(base.TestCase):
    def setUp(self):
        super(TestProfilerRequestContextSerializer, self).setUp()
        self.ser = rpc.ProfilerRequestContextSerializer(mock.Mock())

    @mock.patch('zun.common.rpc.profiler')
    def test_serialize_context(self, mock_profiler):
        prof = mock_profiler.get.return_value
        prof.hmac_key = 'swordfish'
        prof.get_base_id.return_value = 'baseid'
        prof.get_id.return_value = 'parentid'

        context = mock.Mock()
        context.to_dict.return_value = {'project_id': 'test'}

        self.assertEqual({
            'project_id': 'test',
            'trace_info': {
                'hmac_key': 'swordfish',
                'base_id': 'baseid',
                'parent_id': 'parentid'
            }
        }, self.ser.serialize_context(context))

    @mock.patch('zun.common.rpc.profiler')
    def test_deserialize_context(self, mock_profiler):
        serialized = {'project_id': 'test',
                      'trace_info': {
                          'hmac_key': 'swordfish',
                          'base_id': 'baseid',
                          'parent_id': 'parentid'}}

        context = self.ser.deserialize_context(serialized)

        self.assertEqual('test', context.project_id)
        mock_profiler.init.assert_called_once_with(
            hmac_key='swordfish', base_id='baseid', parent_id='parentid')
